# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import glob
import hashlib
import os

from twitter.common.dirutil import safe_mkdir, safe_delete

from twitter.pants import Config, TaskError
from twitter.pants.java import SubprocessExecutor

from .ivy import Bootstrapper, Ivy


def profile_classpath(profile, java_executor=None, config=None):
  """Loads the given profile's classpath if necessary and returns a tuple of
  (updated: bool, classpath: [string]).

  :param string profile: Name of a jvm tool profile.
  :param java_executor: An optional java executor.
  :type java_executor: :class:`twitter.pants.java.Executor`
  :param config: An optional site config.
  :type config: :class:`twitter.pants.base.Config`
  :param string ivy_jar: Optional path to a self-contained ivy jar.
  :param string ivy_settings: Optional path to an ivysettings.xml file.
  """
  # TODO(John Sirois): rework profiles to use the IvyResolve infrastructure used for all other
  # classpath resolution needs.

  # The mechanism used is to store the resolved classpath in a *.checked file whose name contains
  # the hash of the ivy file contents.  In the fast path this allows obtaining the previously
  # resolved classpath by hashing the ivy file, and reading the *.checked file that suggests for
  # the classpath contained within.  When the *.checked for the hash does not exist, this implies
  # the ivy file changed and we need to re-resolve.  In this slow-path case we clean up all the
  # old *.checked files for this profile.

  config = config or Config.load()
  profile_dir = config.get('ivy-profiles', 'workdir')
  ivy_xml = os.path.join(profile_dir, '%s.ivy.xml' % profile)
  if not os.path.exists(ivy_xml):
    raise TaskError('The ivy.xml to configure the %s tool classpath at %s '
                    'is missing.' % (profile, ivy_xml))

  digest = hashlib.sha1()
  with open(ivy_xml) as fp:
    digest.update(fp.read())

  profile_libdir = os.path.join(profile_dir, '%s.libs' % profile)
  profile_check = '%s.%s.checked' % (profile_libdir, digest.hexdigest())
  updated = False
  if not os.path.exists(profile_check):
    safe_mkdir(profile_libdir, clean=True)
    for path in glob.glob('%s*.checked' % profile_libdir):
      safe_delete(path)

    java_executor = java_executor or SubprocessExecutor()
    try:
      ivy = Bootstrapper.default_ivy(java_executor)
      ivy.execute([
        '-ivy', ivy_xml,

        # TODO(John Sirois): this pattern omits an [organisation]- prefix to satisfy IDEA jar naming
        # needs for scala - isolate this hack to idea.py where it belongs
        '-retrieve', '%s/[artifact]-[revision](-[classifier]).[ext]' % profile_libdir,

        # TODO(John Sirois): just use -cachepath and let ivy generate the classpath file contents
        # directly

        '-symlink',
        '-types', 'jar', 'bundle',
        '-confs', 'default'
      ])
    except (Bootstrapper.Error, Ivy.Error) as e:
      raise TaskError('Failed to load profile %s! %s' % (profile_check, e))

    with open(profile_check, 'w') as cp:
      for jar in os.listdir(profile_libdir):
        cp.write('%s\n' % os.path.join(profile_libdir, jar))
    updated = True

  with open(profile_check) as cp:
    return updated, map(lambda entry: entry.strip(), cp.readlines())



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

import inspect
import os
import sys
import traceback

from contextlib import contextmanager
from optparse import Option, OptionParser

try:
  from colors import cyan, green, red, yellow
except ImportError:
  turn_off_colored_logging = True
else:
  turn_off_colored_logging = False

from functools import wraps

from twitter.common import log
from twitter.common.log.options import LogOptions
from twitter.common.collections import OrderedSet
from twitter.common.dirutil import safe_mkdir, safe_rmtree
from twitter.common.lang import Compatibility
from twitter.pants import get_buildroot, goal, group, has_sources, is_apt
from twitter.pants.base import (
    Address,
    BuildFile,
    Config,
    ParseContext,
    Target,
    TargetDefinitionException)
from twitter.pants.base.rcfile import RcFile
from twitter.pants.buildtimestats import BuildTimeStats
from twitter.pants.commands import Command
from twitter.pants.engine import Engine, GroupEngine
from twitter.pants.goal import Context, GoalError, Phase
from twitter.pants.tasks import Task, TaskError
from twitter.pants.tasks.targets_help import TargetsHelp

StringIO = Compatibility.StringIO


class List(Task):
  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    option_group.add_option(mkflag("all"), dest="goal_list_all", default=False, action="store_true",
                            help="[%default] List all goals even if no description is available.")

  def execute(self, targets):
    print('Installed goals:')
    documented_rows = []
    undocumented = []
    max_width = 0
    for phase, _ in Phase.all():
      if phase.description:
        documented_rows.append((phase.name, phase.description))
        max_width = max(max_width, len(phase.name))
      elif self.context.options.goal_list_all:
        undocumented.append(phase.name)
    for name, description in documented_rows:
      print('  %s: %s' % (name.rjust(max_width), description))
    if undocumented:
      print('\nUndocumented goals:\n  %s' % ' '.join(undocumented))


goal(name='goals', action=List).install().with_description('List all documented goals.')


goal(name='targets', action=TargetsHelp).install().with_description('List all target types.')


class Help(Task):
  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    default = None
    if len(args) > 1 and (not args[1].startswith('-')):
      default = args[1]
      del args[1]
    option_group.add_option(mkflag("goal"), dest="help_goal", default=default)

  def execute(self, targets):
    goal = self.context.options.help_goal
    if goal is None:
      return self.list_goals('You must supply a goal name to provide help for.')
    phase = Phase(goal)
    if not phase.goals():
      return self.list_goals('Goal %s is unknown.' % goal)

    parser = OptionParser()
    parser.set_usage('%s goal %s ([target]...)' % (sys.argv[0], goal))
    parser.epilog = phase.description
    Goal.add_global_options(parser)
    Phase.setup_parser(parser, [], [phase])
    parser.parse_args(['--help'])

  def list_goals(self, message):
    print(message)
    print()
    return Goal.execute(self.context, 'goals')

goal(name='help', action=Help).install().with_description('Provide help for the specified goal.')


class SpecParser(object):
  """Parses goal target specs; either simple target addresses or else sibling (:) or descendant
  (::) selector forms
  """

  def __init__(self, root_dir):
    self._root_dir = root_dir

  def _get_dir(self, spec):
    path = spec.split(':', 1)[0]
    if os.path.isdir(path):
      return path
    else:
      if os.path.isfile(path):
        return os.path.dirname(path)
      else:
        return spec

  def _parse_addresses(self, spec):
    if spec.endswith('::'):
      dir = self._get_dir(spec[:-len('::')])
      for buildfile in BuildFile.scan_buildfiles(self._root_dir, os.path.join(self._root_dir, dir)):
        for address in Target.get_all_addresses(buildfile):
          yield address
    elif spec.endswith(':'):
      dir = self._get_dir(spec[:-len(':')])
      for address in Target.get_all_addresses(BuildFile(self._root_dir, dir)):
        yield address
    else:
      yield Address.parse(self._root_dir, spec)

  def parse(self, spec):
    """Parses the given target spec into one or more targets.

    Returns a generator of target, address pairs in which the target may be None if the address
    points to a non-existent target.
    """
    for address in self._parse_addresses(spec):
      target = Target.get(address)
      yield target, address


class Goal(Command):
  """Lists installed goals or else executes a named goal."""

  __command__ = 'goal'

  GLOBAL_OPTIONS = [
    Option("-t", "--timeout", dest="conn_timeout", type='int',
           default=Config.load().getdefault('connection_timeout'),
           help="Number of seconds to wait for http connections."),
    Option("-x", "--time", action="store_true", dest="time", default=False,
           help="Times goal phases and outputs a report."),
    Option("-e", "--explain", action="store_true", dest="explain", default=False,
           help="Explain the execution of goals."),
    Option("--force-stats-upload", dest="force_upload", action="store_true", default=False,
           help="[%default] Forces pants runtime stats upload after every pants runs"),
    Option("-v", "--log", action="store_true", dest="log", default=False,
           help="[%default] Logs extra build output."),
    Option("-d", "--logdir", dest="logdir",
           help="[%default] Forks logs to files under this directory."),
    Option("-l", "--level", dest="log_level", type="choice", choices=['debug', 'info', 'warn'],
           help="[info] Sets the logging level to one of 'debug', 'info' or 'warn', implies -v "
                  "if set."),
    Option("--no-colors", dest="no_color", action="store_true", default=turn_off_colored_logging,
           help="Do not colorize log messages."),
    Option("--all", dest="target_directory", action="append",
           help="DEPRECATED: Use [dir]: with no flag in a normal target position on the command "
                  "line. (Adds all targets found in the given directory's BUILD file. Can be "
                  "specified more than once.)"),
    Option("--all-recursive", dest="recursive_directory", action="append",
           help="DEPRECATED: Use [dir]:: with no flag in a normal target position on the command "
                  "line. (Adds all targets found recursively under the given directory. Can be "
                  "specified more than once to add more than one root target directory to scan.)")
  ]

  @staticmethod
  def add_global_options(parser):
    for option in Goal.GLOBAL_OPTIONS:
      parser.add_option(option)

  @staticmethod
  def parse_args(args):
    goals = OrderedSet()
    specs = OrderedSet()
    help = False
    explicit_multi = False

    def is_spec(spec):
      return os.sep in spec or ':' in spec

    for i, arg in enumerate(args):
      help = help or 'help' == arg
      if not arg.startswith('-'):
        specs.add(arg) if is_spec(arg) else goals.add(arg)
      elif '--' == arg:
        if specs:
          raise GoalError('Cannot intermix targets with goals when using --. Targets should '
                          'appear on the right')
        explicit_multi = True
        del args[i]
        break

    if explicit_multi:
      spec_offset = len(goals) + 1 if help else len(goals)
      specs.update(arg for arg in args[spec_offset:] if not arg.startswith('-'))

    return goals, specs

  @staticmethod
  def execute(context, *names):
    parser = OptionParser()
    phases = [Phase(name) for name in names]
    Phase.setup_parser(parser, [], phases)
    options, _ = parser.parse_args([])
    context = Context(context.config, options, context.target_roots, log=context.log)
    return Goal._execute(context, phases, print_timing=False, force_stats_upload=False)

  @staticmethod
  def _execute(context, phases, print_timing, force_stats_upload):
    build_stats = None
    if context.config.getdefault('stats_collection', bool, default=False):
      user = context.config.getdefault('user')
      build_stats = BuildTimeStats(user, force_stats_upload)
    engine = GroupEngine(print_timing=print_timing, build_stats=build_stats)
    return engine.execute(context, phases)

  # TODO(John Sirois): revisit wholesale locking when we move py support into pants new
  @classmethod
  def serialized(cls):
    return True

  def __init__(self, root_dir, parser, args):
    self.targets = []
    Command.__init__(self, root_dir, parser, args)

  @contextmanager
  def check_errors(self, banner):
    errors = {}
    def error(key, include_traceback=False):
      exc_type, exc_value, _ = sys.exc_info()
      msg = StringIO()
      if include_traceback:
        frame = inspect.trace()[-1]
        filename = frame[1]
        lineno = frame[2]
        funcname = frame[3]
        code = ''.join(frame[4])
        traceback.print_list([(filename, lineno, funcname, code)], file=msg)
      if exc_type:
        msg.write(''.join(traceback.format_exception_only(exc_type, exc_value)))
      errors[key] = msg.getvalue()
      sys.exc_clear()

    yield error

    if errors:
      msg = StringIO()
      msg.write(banner)
      invalid_keys = [key for key, exc in errors.items() if not exc]
      if invalid_keys:
        msg.write('\n  %s' % '\n  '.join(invalid_keys))
      for key, exc in errors.items():
        if exc:
          msg.write('\n  %s =>\n    %s' % (key, '\n      '.join(exc.splitlines())))
      self.error(msg.getvalue())

  def setup_parser(self, parser, args):
    self.config = Config.load()

    Goal.add_global_options(parser)

    # We support attempting zero or more goals.  Multiple goals must be delimited from further
    # options and non goal args with a '--'.  The key permutations we need to support:
    # ./pants goal => goals
    # ./pants goal goals => goals
    # ./pants goal compile src/java/... => compile
    # ./pants goal compile -x src/java/... => compile
    # ./pants goal compile src/java/... -x => compile
    # ./pants goal compile run -- src/java/... => compile, run
    # ./pants goal compile run -- src/java/... -x => compile, run
    # ./pants goal compile run -- -x src/java/... => compile, run

    if not args:
      args.append('goals')

    if len(args) == 1 and args[0] in set(['-h', '--help', 'help']):
      def format_usage(usages):
        left_colwidth = 0
        for left, right in usages:
          left_colwidth = max(left_colwidth, len(left))
        lines = []
        for left, right in usages:
          lines.append('  %s%s%s' % (left, ' ' * (left_colwidth - len(left) + 1), right))
        return '\n'.join(lines)

      usages = [
        ("%prog goal goals ([spec]...)", Phase('goals').description),
        ("%prog goal help [goal] ([spec]...)", Phase('help').description),
        ("%prog goal [goal] [spec]...", "Attempt goal against one or more targets."),
        ("%prog goal [goal] ([goal]...) -- [spec]...", "Attempts all the specified goals."),
      ]
      parser.set_usage("\n%s" % format_usage(usages))
      parser.epilog = ("Either lists all installed goals, provides extra help for a goal or else "
                       "attempts to achieve the specified goal for the listed targets." """
                       Note that target specs accept two special forms:
                         [dir]:  to include all targets in the specified directory
                         [dir]:: to include all targets found in all BUILD files recursively under
                                 the directory""")

      parser.print_help()
      sys.exit(0)
    else:
      goals, specs = Goal.parse_args(args)

      # Bootstrap goals by loading any configured bootstrap BUILD files
      with self.check_errors('The following bootstrap_buildfiles cannot be loaded:') as error:
        for path in self.config.getlist('goals', 'bootstrap_buildfiles', default=[]):
          try:
            buildfile = BuildFile(get_buildroot(), os.path.relpath(path, get_buildroot()))
            ParseContext(buildfile).parse()
          except (TypeError, ImportError, TaskError, GoalError):
            error(path, include_traceback=True)
          except (IOError, SyntaxError):
            error(path)

      # Bootstrap user goals by loading any BUILD files implied by targets
      spec_parser = SpecParser(self.root_dir)
      with self.check_errors('The following targets could not be loaded:') as error:
        for spec in specs:
          try:
            for target, address in spec_parser.parse(spec):
              if target:
                self.targets.append(target)
                # Force early BUILD file loading if this target is an alias that expands to others.
                unused = list(target.resolve())
              else:
                siblings = Target.get_all_addresses(address.buildfile)
                prompt = 'did you mean' if len(siblings) == 1 else 'maybe you meant one of these'
                error('%s => %s?:\n    %s' % (address, prompt,
                                              '\n    '.join(str(a) for a in siblings)))
          except (TypeError, ImportError, TaskError, GoalError):
            error(spec, include_traceback=True)
          except (IOError, SyntaxError, TargetDefinitionException):
            error(spec)

      self.phases = [Phase(goal) for goal in goals]

      rcfiles = self.config.getdefault('rcfiles', type=list, default=[])
      if rcfiles:
        rcfile = RcFile(rcfiles, default_prepend=False, process_default=True)

        # Break down the goals specified on the command line to the full set that will be run so we
        # can apply default flags to inner goal nodes.  Also break down goals by Task subclass and
        # register the task class hierarchy fully qualified names so we can apply defaults to
        # baseclasses.

        sections = OrderedSet()
        for phase in Engine.execution_order(self.phases):
          for goal in phase.goals():
            sections.add(goal.name)
            for clazz in goal.task_type.mro():
              if clazz == Task:
                break
              sections.add('%s.%s' % (clazz.__module__, clazz.__name__))

        augmented_args = rcfile.apply_defaults(sections, args)
        if augmented_args != args:
          del args[:]
          args.extend(augmented_args)
          print("(using pantsrc expansion: pants goal %s)" % ' '.join(augmented_args),
                file=sys.stderr)

      Phase.setup_parser(parser, args, self.phases)

  def run(self, lock):
    logger = None
    if self.options.log or self.options.log_level:
      LogOptions.set_stderr_log_level((self.options.log_level or 'info').upper())
      logdir = self.options.logdir or self.config.get('goals', 'logdir', default=None)
      if logdir:
        safe_mkdir(logdir)
        LogOptions.set_log_dir(logdir)
        log.init('goals')
      else:
        log.init()
      logger = log

      if not self.options.no_color:
        def colorwrapper(func, clrname):
          @wraps(func)
          def wrapper(msg, *args, **kwargs):
            return func(clrname(msg), *args, **kwargs)
          return wrapper

        log.info = colorwrapper(log.info, green)
        log.warn = colorwrapper(log.warn, yellow)
        log.error = colorwrapper(log.error, red)
        log.debug = colorwrapper(log.debug, cyan)

    if self.options.recursive_directory:
      log.warn('--all-recursive is deprecated, use a target spec with the form [dir]:: instead')
      for dir in self.options.recursive_directory:
        self.add_target_recursive(dir)

    if self.options.target_directory:
      log.warn('--all is deprecated, use a target spec with the form [dir]: instead')
      for dir in self.options.target_directory:
        self.add_target_directory(dir)

    buildtime_stats = BuildTimeStats(self.config.getdefault("user"),
                                     force_upload=self.options.force_upload)

    context = Context(self.config, self.options, self.targets, lock=lock, log=logger,
                      buildtime_stats=buildtime_stats)

    unknown = []
    for phase in self.phases:
      if not phase.goals():
        unknown.append(phase)

    if unknown:
      print('Unknown goal(s): %s\n' % ' '.join(phase.name for phase in unknown))
      Goal.execute(context, 'goals')
      return 1

    if logger:
      logger.debug('Operating on targets: %s' % self.targets)

    return Goal._execute(context, self.phases, print_timing=self.options.time,
                         force_stats_upload=self.options.force_upload)


# Install all default pants provided goals
from twitter.pants import junit_tests
from twitter.pants.targets import (
  Benchmark,
  JavaLibrary,
  JvmBinary,
  ScalacPlugin,
  ScalaLibrary,
  ScalaTests)
from twitter.pants.tasks.antlr_gen import AntlrGen
from twitter.pants.tasks.benchmark_run import BenchmarkRun
from twitter.pants.tasks.binary_create import BinaryCreate
from twitter.pants.tasks.builddictionary import BuildBuildDictionary
from twitter.pants.tasks.bundle_create import BundleCreate
from twitter.pants.tasks.checkstyle import Checkstyle
from twitter.pants.tasks.check_published_deps import CheckPublishedDeps
from twitter.pants.tasks.detect_duplicates import DuplicateDetector
from twitter.pants.tasks.extract import Extract
from twitter.pants.tasks.filedeps import FileDeps
from twitter.pants.tasks.idl_extract import IdlExtract
from twitter.pants.tasks.idl_resolve import IdlResolve
from twitter.pants.tasks.ivy_resolve import IvyResolve
from twitter.pants.tasks.jar_create import JarCreate
from twitter.pants.tasks.java_compile import JavaCompile
from twitter.pants.tasks.javadoc_gen import JavadocGen
from twitter.pants.tasks.scaladoc_gen import ScaladocGen
from twitter.pants.tasks.junit_run import JUnitRun
from twitter.pants.tasks.jvm_run import JvmRun
from twitter.pants.tasks.markdown_to_html import MarkdownToHtml
from twitter.pants.tasks.nailgun_task import NailgunTask
from twitter.pants.tasks.listtargets import ListTargets
from twitter.pants.tasks.pathdeps import PathDeps
from twitter.pants.tasks.prepare_resources import PrepareResources
from twitter.pants.tasks.protobuf_gen import ProtobufGen
from twitter.pants.tasks.scala_compile import ScalaCompile
from twitter.pants.tasks.scala_repl import ScalaRepl
from twitter.pants.tasks.specs_run import SpecsRun
from twitter.pants.tasks.thrift_gen import ThriftGen
from twitter.pants.tasks.scrooge_gen import ScroogeGen


class Invalidator(Task):
  def execute(self, targets):
    self.invalidate(all_=True)
goal(name='invalidate', action=Invalidator).install().with_description('Invalidate all caches')


goal(
  name='clean-all',
  action=lambda ctx: safe_rmtree(ctx.config.getdefault('pants_workdir')),
  dependencies=['invalidate']
).install().with_description('Cleans all intermediate build output')


if NailgunTask.killall:
  class NailgunKillall(Task):
    @classmethod
    def setup_parser(cls, option_group, args, mkflag):
      option_group.add_option(mkflag("everywhere"), dest="ng_killall_evywhere",
                              default=False, action="store_true",
                              help="[%default] Kill all nailguns servers launched by pants for "
                                   "all workspaces on the system.")

    def execute(self, targets):
      if NailgunTask.killall:
        NailgunTask.killall(logger=self.context.log.info,
                            everywhere=self.context.options.ng_killall_evywhere)

  ng_killall = goal(name='ng-killall', action=NailgunKillall)
  ng_killall.install().with_description('Kill any running nailgun servers spawned by pants.')

  ng_killall.install('clean-all', first=True)


# TODO(John Sirois): Resolve eggs
goal(
  name='ivy',
  action=IvyResolve,
  dependencies=['gen']
).install('resolve').with_description('Resolves jar dependencies and produces dependency reports.')

goal(
  name='idl',
  action=IdlResolve,
).install('resolve-idl').with_description('Resolves idl jar dependencies.')

goal(
  name='extract',
  action=Extract,
).install('resolve-idl')

goal(
  name='idl-extract',
  action=IdlExtract,
).install('resolve-idl')

# TODO(John Sirois): gen attempted as the sole Goal should gen for all known gen types but
# recognize flags to narrow the gen set
goal(name='thrift', action=ThriftGen,
  dependencies=['resolve-idl']).install('gen').with_description('Generate code.')
goal(name='scrooge', action=ScroogeGen,
  dependencies=['resolve-idl']).install('gen')
goal(name='protoc', action=ProtobufGen,
  dependencies=['resolve-idl']).install('gen')
goal(name='antlr', action=AntlrGen,
  dependencies=['resolve-idl']).install('gen')

goal(
  name='checkstyle',
  action=Checkstyle,
  dependencies=['gen', 'resolve']
).install().with_description('Run checkstyle against java source code.')

# TODO(John Sirois): These group predicates could simplify to simple has_sources checks except for
# the fact that sometimes 'aggregator' targets with no sources serve as a dependency link in the
# wild.  Consider stomping this practice out and simplifying these predicates.

def is_java(target):
  return (isinstance(target, JavaLibrary)
          or (isinstance(target, (JvmBinary, junit_tests, Benchmark))
              and has_sources(target, '.java')))


def is_scala(target):
  return (isinstance(target, (ScalaLibrary, ScalaTests, ScalacPlugin))
          or (isinstance(target, (JvmBinary, junit_tests, Benchmark))
              and has_sources(target, '.scala')))


goal(name='scalac',
     action=ScalaCompile,
     group=group('jvm', is_scala),
     dependencies=['gen', 'resolve']).install('compile').with_description(
       'Compile both generated and checked in code.'
     )
goal(name='apt',
     action=JavaCompile,
     group=group('jvm', is_apt),
     dependencies=['gen', 'resolve']).install('compile')
goal(name='javac',
     action=JavaCompile,
     group=group('jvm', is_java),
     dependencies=['gen', 'resolve']).install('compile')


goal(name='prepare', action=PrepareResources).install('resources')


# TODO(John Sirois): pydoc also
goal(name='javadoc',
     action=JavadocGen,
     dependencies=['compile']).install('doc').with_description('Create documentation.')
goal(name='scaladoc',
     action=ScaladocGen,
     dependencies=['compile']).install('doc')


if MarkdownToHtml.AVAILABLE:
  goal(name='markdown',
       action=MarkdownToHtml
  ).install('markdown').with_description('Generate html from markdown docs.')


class ScaladocJarShim(ScaladocGen):
  def __init__(self, context, output_dir=None, confs=None):
    super(ScaladocJarShim, self).__init__(context,
                                          output_dir=output_dir,
                                          confs=confs,
                                          active=False)


class JavadocJarShim(JavadocGen):
  def __init__(self, context, output_dir=None, confs=None):
    super(JavadocJarShim, self).__init__(context,
                                         output_dir=output_dir,
                                         confs=confs,
                                         active=False)


class JarCreateGoal(JarCreate):
  def __init__(self, context):
    super(JarCreateGoal, self).__init__(context, False)

goal(name='javadoc_publish',
     action=JavadocJarShim).install('jar')
goal(name='scaladoc_publish',
     action=ScaladocJarShim).install('jar')
goal(name='jar',
     action=JarCreateGoal,
     dependencies=['compile', 'resources']).install('jar').with_description('Create one or more jars.')
goal(name='check_published_deps',
     action=CheckPublishedDeps
).install('check_published_deps').with_description(
  'Find references to outdated artifacts published from this BUILD tree.')

goal(name='junit',
     action=JUnitRun,
     dependencies=['compile', 'resources']).install('test').with_description('Test compiled code.')
goal(name='specs',
     action=SpecsRun,
     dependencies=['compile', 'resources']).install('test')

goal(name='bench',
     action=BenchmarkRun,
     dependencies=['compile', 'resources']).install('bench')

# TODO(John Sirois): Create pex's in binary phase
goal(
  name='binary',
  action=BinaryCreate,
  dependencies=['jar']
).install().with_description('Create a jvm binary jar.')
goal(
  name='dup',
  action=DuplicateDetector,
).install('binary')
goal(
  name='bundle',
  action=BundleCreate,
  dependencies=['binary']
).install().with_description('Create an application bundle from binary targets.')

goal(
  name='detect-duplicates',
  action=DuplicateDetector,
  dependencies=['jar']
).install().with_description('Detect duplicate classes and resources on the classpath.')

goal(
  name='jvm-run',
  action=JvmRun,
  dependencies=['compile', 'resources']
).install('run').with_description('Run a (currently JVM only) binary target.')

goal(
  name='scala-repl',
  action=ScalaRepl,
  dependencies=['compile', 'resources']
).install('repl').with_description(
  'Run a (currently Scala only) REPL with the classpath set according to the targets.')

goal(
  name='filedeps',
  action=FileDeps
).install('filedeps').with_description('Print out a list of all files the target depends on')

goal(
  name='pathdeps',
  action=PathDeps
).install('pathdeps').with_description(
  'Print out a list of all paths containing build files the target depends on')

goal(
  name='list',
  action=ListTargets
).install('list').with_description('List available BUILD targets.')

goal(  # Generate the BUILD Dictionary
  name='builddict',
  action=BuildBuildDictionary
).install('builddict')

from twitter.pants.tasks.idea_gen import IdeaGen

goal(
  name='idea',
  action=IdeaGen,
  dependencies=['jar']
).install().with_description('Create an IntelliJ IDEA project from the given targets.')


from twitter.pants.tasks.eclipse_gen import EclipseGen

goal(
  name='eclipse',
  action=EclipseGen,
  dependencies=['jar']
).install().with_description('Create an Eclipse project from the given targets.')


from twitter.pants.tasks.python.setup import SetupPythonEnvironment

goal(
  name='python-setup',
  action=SetupPythonEnvironment,
).install('setup').with_description(
"Setup the target's build environment.")


from twitter.pants.tasks.dependees import ReverseDepmap

goal(
  name='dependees',
  action=ReverseDepmap
).install().with_description('Print a reverse dependency mapping for the given targets')


from twitter.pants.tasks.depmap import Depmap

goal(
  name='depmap',
  action=Depmap
).install().with_description('Generates either a textual dependency tree or a graphviz'
                             ' digraph dotfile for the dependency set of a target')


from twitter.pants.tasks.dependencies import Dependencies

goal(
  name='dependencies',
  action=Dependencies
).install().with_description('Extract textual infomation about the dependencies of a target')


from twitter.pants.tasks.filemap import Filemap

goal(
  name='filemap',
  action=Filemap
).install().with_description('Outputs a mapping from source file to'
                             ' the target that owns the source file')


from twitter.pants.tasks.minimal_cover import MinimalCover

goal(
  name='minimize',
  action=MinimalCover
).install().with_description('Print the minimal cover of the given targets.')


from twitter.pants.tasks.filter import Filter

goal(
  name='filter',
  action=Filter
).install().with_description('Filter the input targets based on various criteria.')


from twitter.pants.tasks.sorttargets import SortTargets

goal(
  name='sort',
  action=SortTargets
).install().with_description('Topologically sort the input targets.')


from twitter.pants.tasks.roots import ListRoots

goal(
  name='roots',
  action=ListRoots,
).install('roots').with_description("Prints the source roots and associated target types defined in the repo.")



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

try:
  import ConfigParser
except ImportError:
  import configparser as ConfigParser

import os
import getpass

from twitter.pants import get_buildroot


class Config(object):
  """
    Encapsulates ini-style config file loading and access additionally supporting recursive variable
    substitution using standard python format strings, ie: %(var_name)s will be replaced with the
    value of var_name.
  """

  DEFAULT_SECTION = ConfigParser.DEFAULTSECT

  class ConfigError(Exception): pass

  @staticmethod
  def load(configpath=None, defaults=None):
    """
      Loads a Config from the given path, by default the path to the pants.ini file in the current
      build root directory.  Any defaults supplied will act as if specified in the loaded config
      file's DEFAULT section.  The 'buildroot', invoking 'user' and invoking user's 'homedir' are
      automatically defaulted.
    """
    configpath = configpath or os.path.join(get_buildroot(), 'pants.ini')
    parser = Config.create_parser(defaults=defaults)
    with open(configpath) as ini:
      parser.readfp(ini)
    return Config(parser)

  @staticmethod
  def create_parser(defaults=None):
    """
      Creates a config parser that supports %([key-name])s value substitution.  Any defaults
      supplied will act as if specified in the loaded config file's DEFAULT section and be available
      for substitutions.  The 'buildroot', invoking 'user' and invoking user's 'homedir' are
      automatically defaulted.
    """
    standard_defaults = dict(
      buildroot=get_buildroot(),
      homedir=os.path.expanduser('~'),
      user=getpass.getuser(),
      pants_workdir=os.path.join(get_buildroot(), '.pants.d'),
      pants_supportdir=os.path.join(get_buildroot(), 'build-support'),
      pants_distdir=os.path.join(get_buildroot(), 'dist')
    )
    if defaults:
      standard_defaults.update(defaults)
    return ConfigParser.SafeConfigParser(standard_defaults)

  def __init__(self, configparser):
    self.configparser = configparser

  def getbool(self, section, option, default=None):
    """Equivalent to calling get with expected type string"""
    return self.get(section, option, type=bool, default=default)

  def getint(self, section, option, default=None):
    """Equivalent to calling get with expected type int"""
    return self.get(section, option, type=int, default=default)

  def getfloat(self, section, option, default=None):
    """Equivalent to calling get with expected type float"""
    return self.get(section, option, type=float, default=default)

  def getlist(self, section, option, default=None):
    """Equivalent to calling get with expected type list"""
    return self.get(section, option, type=list, default=default)

  def getdict(self, section, option, default=None):
    """Equivalent to calling get with expected type dict"""
    return self.get(section, option, type=dict, default=default)

  def getdefault(self, option, type=str, default=None):
    """
      Retrieves option from the DEFAULT section if it exists and attempts to parse it as type.
      If there is no definition found, the default value supplied is returned.
    """
    return self.get(Config.DEFAULT_SECTION, option, type, default=default)

  def get(self, section, option, type=str, default=None):
    """
      Retrieves option from the specified section if it exists and attempts to parse it as type.
      If the specified section is missing a definition for the option, the value is looked up in the
      DEFAULT section.  If there is still no definition found, the default value supplied is
      returned.
    """
    return self._getinstance(section, option, type, default=default)

  def get_required(self, section, option, type=str):
    """Retrieves option from the specified section and attempts to parse it as type.
    If the specified section is missing a definition for the option, the value is
    looked up in the DEFAULT section. If there is still no definition found,
    a `ConfigError` is raised.

    :param string section: Section to lookup the option in, before looking in DEFAULT.
    :param string option: Option to retrieve.
    :param type: Type to retrieve the option as.
    :returns: The option as the specified type.
    :raises: :class:`twitter.pants.base.config.Config.ConfigError` if option is not found.
    """
    val = self.get(section, option, type=type)
    if val is None:
      raise Config.ConfigError('Required option %s.%s is not defined.' % (section, option))
    return val

  def has_section(self, section):
    """Return whether or not this config has the section."""
    return self.configparser.has_section(section)

  def _getinstance(self, section, option, type, default=None):
    if not self.configparser.has_option(section, option):
      return default
    raw_value = self.configparser.get(section, option)
    if issubclass(type, str):
      return raw_value

    try:
      parsed_value = eval(raw_value, {}, {})
    except SyntaxError as e:
      raise Config.ConfigError('No valid %s for %s.%s: %s\n%s' % (
        type.__name__, section, option, raw_value, e))

    if not isinstance(parsed_value, type):
      raise Config.ConfigError('No valid %s for %s.%s: %s' % (
        type.__name__, section, option, raw_value))

    return parsed_value



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os

from twitter.pants import has_sources, is_jvm
from twitter.pants.tasks.jvmdoc_gen import Jvmdoc, JvmdocGen


javadoc = Jvmdoc(tool_name='javadoc', product_type='javadoc')


def is_java(target):
  return has_sources(target, '.java')


class JavadocGen(JvmdocGen):
  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    cls.generate_setup_parser(option_group, args, mkflag, javadoc)

  def __init__(self, context, output_dir=None, confs=None, active=True):
    super(JavadocGen, self).__init__(context, javadoc, output_dir, confs, active)

  def execute(self, targets):
    self.generate_execute(targets, is_java, create_javadoc_command)


def create_javadoc_command(classpath, gendir, *targets):
  sources = []
  for target in targets:
    sources.extend(os.path.join(target.target_base, source) for source in target.sources)

  if not sources:
    return None

  # TODO(John Sirois): try com.sun.tools.javadoc.Main via ng
  command = [
    'javadoc',
    '-quiet',
    '-encoding', 'UTF-8',
    '-notimestamp',
    '-use',
    '-classpath', ':'.join(classpath),
    '-d', gendir,
  ]

  # Always provide external linking for java API
  offlinelinks = set(['http://download.oracle.com/javase/6/docs/api/'])

  def link(target):
    for jar in target.jar_dependencies:
      if jar.apidocs:
        offlinelinks.add(jar.apidocs)
  for target in targets:
    target.walk(link, is_jvm)

  for link in offlinelinks:
    command.extend(['-linkoffline', link, link])

  command.extend(sources)
  return command



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

import inspect
import sys

from twitter.common.collections import OrderedSet
from twitter.pants.base import BuildFile, Target


class Command(object):
  """Baseclass for all pants subcommands."""

  _commands = {}

  @staticmethod
  def get_command(name):
    return Command._commands.get(name, None)

  @staticmethod
  def all_commands():
    return Command._commands.keys()

  @staticmethod
  def register_commands():
    """Register all 'Command's from all modules in the current directory."""
    import pkgutil
    for _, mod, ispkg in pkgutil.iter_modules(__path__):
      if ispkg: continue
      fq_module = '.'.join([__name__, mod])
      __import__(fq_module)
      for (_, kls) in inspect.getmembers(sys.modules[fq_module], inspect.isclass):
        if issubclass(kls, Command):
          command_name = kls.__dict__.get('__command__', None)
          if command_name:
            Command._commands[command_name] = kls

  @staticmethod
  def scan_addresses(root_dir, base_path = None):
    """Parses all targets available in BUILD files under base_path and
    returns their addresses.  If no base_path is specified, root_dir is
    assumed to be the base_path"""

    addresses = OrderedSet()
    for buildfile in BuildFile.scan_buildfiles(root_dir, base_path):
      addresses.update(Target.get_all_addresses(buildfile))
    return addresses

  @classmethod
  def serialized(cls):
    return False

  def __init__(self, root_dir, parser, args):
    """root_dir: The root directory of the pants workspace
    parser: an OptionParser
    argv: the subcommand arguments to parse"""

    self.root_dir = root_dir

    # Override the OptionParser's error with more useful output
    def error(message = None):
      if message:
        print(message + '\n')
      parser.print_help()
      parser.exit(status=1)
    parser.error = error
    self.error = error

    self.setup_parser(parser, args)
    self.options, self.args = parser.parse_args(args)
    self.parser = parser

  def setup_parser(self, parser, args):
    """Subclasses should override and confiure the OptionParser to reflect
    the subcommand option and argument requirements.  Upon successful
    construction, subcommands will be able to access self.options and
    self.args."""

    pass

  def error(self, message = None):
    """Reports the error message followed by pants help and then exits."""

  def run(self, lock):
    """Subcommands that are serialized() should override if they need the ability to interact with
    the global command lock.
    The value returned should be an int, 0 indicating success and any other value indicating
    failure."""
    return self.execute()

  def execute(self):
    """Subcommands that do not require serialization should override to perform the command action.
    The value returned should be an int, 0 indicating success and any other value indicating
    failure."""
    raise NotImplementedError('Either run(lock) or execute() must be over-ridden.')



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

__author__ = 'Phil Hom'

import os

from .extract import Extract
from .ivy_resolve import IvyResolve


class IdlResolve(IvyResolve):

  # TODO(John Sirois): rework ivy_resolve to allow idl resolution without subclassing
  #   (via goal dependencies instead)
  def __init__(self, context):
    super(IdlResolve, self).__init__(context, ['idl'])

  # Carve out the custom behavior we need in the base machinery with a few overrides:
  # + don't do a classpath resolve, we only want to resolve individual idl jar closures
  # + don't generate ivy reports, let IvyResolve handle this
  # + only map idl jars: an idl dependency graph should be pure idl jars
  # + warn about idl classpaths that have non-idl jars

  def _extract_classpathdeps(self, targets):
    # IDL files don't contribute to classpaths.
    return set()

  def _generate_ivy_report(self):
    pass

  def _mapfor_typename(self):
    return 'idl_dependencies'

  def _mapto_dir(self):
    return os.path.join(self._work_dir, 'mapped-idls')

  def _map_jar(self, org, name, conf, path):
    # The spec says idl only jars should have an idl classifier and depend only on other such
    # artifacts, but in practice the spec is often not met.  It's useful to allow direct
    # registration of jars containing idl no matter the form of the jar.  We do restrict transitive
    # idl deps however to those that conform with the spec.
    is_idl = Extract.is_registered_jar(org, name) or path.endswith('-idl.jar')
    if not is_idl and super(IdlResolve, self)._map_jar(org, name, conf, path):
      self.context.log.warn('Ignoring invalid idl dependency: %s' % path)
    return is_idl




# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import collections
import contextlib
import multiprocessing
import os
import subprocess

from twitter.common.dirutil import safe_mkdir

from twitter.pants import binary_util, is_codegen
from twitter.pants.tasks import Task, TaskError

Jvmdoc = collections.namedtuple('Jvmdoc', ['tool_name', 'product_type'])

ParserConfig = collections.namedtuple(
                'JvmdocGenParserConfig',
                ['outdir_opt', 'include_codegen_opt', 'transitive_opt', 'open_opt', 'combined_opt',
                 'ignore_failure_opt'])


class JvmdocGen(Task):
  @classmethod
  def setup_parser_config(cls):
    opts = ['%s_%s' % (cls.__name__, opt) for opt in ParserConfig._fields]
    return ParserConfig(*opts)

  @classmethod
  def generate_setup_parser(cls, option_group, args, mkflag, jvmdoc):
    parser_config = cls.setup_parser_config()
    option_group.add_option(
      mkflag('outdir'),
      dest=parser_config.outdir_opt,
      help='Emit %s in this directory.' % jvmdoc.tool_name)

    option_group.add_option(
      mkflag('include-codegen'),
      mkflag('include-codegen', negate=True),
      dest=parser_config.include_codegen_opt,
      default=None,
      action='callback',
      callback=mkflag.set_bool,
      help='[%%default] Create %s for generated code.' % jvmdoc.tool_name)

    option_group.add_option(
      mkflag('transitive'),
      mkflag('transitive', negate=True),
      dest=parser_config.transitive_opt,
      default=True,
      action='callback',
      callback=mkflag.set_bool,
      help='[%%default] Create %s for the transitive closure of internal '
           'targets reachable from the roots specified on the command line.'
           % jvmdoc.tool_name)

    combined_flag = mkflag('combined')
    option_group.add_option(
      combined_flag,
      mkflag('combined', negate=True),
      dest=parser_config.combined_opt,
      default=False,
      action='callback',
      callback=mkflag.set_bool,
      help='[%%default] Generate %s for all targets combined instead of '
           'each target individually.'
           % jvmdoc.tool_name)

    option_group.add_option(
      mkflag('open'),
      mkflag('open', negate=True),
      dest=parser_config.open_opt,
      default=False,
      action='callback',
      callback=mkflag.set_bool,
      help='[%%default] Attempt to open the generated %s in a browser '
           '(implies %s).' % (jvmdoc.tool_name, combined_flag))

    option_group.add_option(
      mkflag('ignore-failure'),
      mkflag('ignore-failure', negate=True),
      dest=parser_config.ignore_failure_opt,
      default=False,
      action='callback',
      callback=mkflag.set_bool,
      help='Specifies that %s errors should not cause build errors'
           % jvmdoc.tool_name)

  def __init__(self, context, jvmdoc, output_dir, confs, active):
    def getattr_options(option):
      return getattr(context.options, option)

    super(JvmdocGen, self).__init__(context)

    self._jvmdoc = jvmdoc
    jvmdoc_tool_name = self._jvmdoc.tool_name

    config_section = '%s-gen' % jvmdoc_tool_name
    parser_config = self.setup_parser_config()

    pants_workdir = context.config.getdefault('pants_workdir')
    self._output_dir = (
      output_dir
      or getattr_options(parser_config.outdir_opt)
      or context.config.get(config_section,
                            'workdir',
                            default=os.path.join(pants_workdir, jvmdoc_tool_name))
    )

    flagged_codegen = getattr_options(parser_config.include_codegen_opt)
    self._include_codegen = (flagged_codegen if flagged_codegen is not None
                             else context.config.getbool(config_section, 'include_codegen',
                                                         default=False))

    self.transitive = getattr_options(parser_config.transitive_opt)
    self.confs = confs or context.config.getlist(config_section, 'confs', default=['default'])
    self.active = active
    self.open = getattr_options(parser_config.open_opt)
    self.combined = self.open or getattr_options(parser_config.combined_opt)
    self.ignore_failure = getattr_options(parser_config.ignore_failure_opt)

  def invalidate_for(self):
    return (self.combined, self.transitive, self._output_dir, self.confs, self._include_codegen)

  def generate_execute(self, targets, language_predicate, create_jvmdoc_command):
    """
    Generate an execute method given a language predicate and command to create documentation

    language_predicate: a function that accepts a target and returns True if the target is of that
                        language
    create_jvmdoc_command: (classpath, directory, *targets) -> command (string) that will generate
                           documentation documentation for targets
    """
    catalog = self.context.products.isrequired(self._jvmdoc.product_type)
    if catalog and self.combined:
      raise TaskError(
          'Cannot provide %s target mappings for combined output' % self._jvmdoc.product_type)
    elif catalog or self.active:
      def docable(target):
        return language_predicate(target) and (self._include_codegen or not is_codegen(target))

      with self.changed(filter(docable, targets)) as changed_targets:
        safe_mkdir(self._output_dir)
        with self.context.state('classpath', []) as cp:
          classpath = [jar for conf, jar in cp if conf in self.confs]

          def find_jvmdoc_targets():
            if self.transitive:
              return changed_targets
            else:
              return set(changed_targets).intersection(set(self.context.target_roots))

          jvmdoc_targets = list(filter(docable, find_jvmdoc_targets()))
          if self.combined:
            self._generate_combined(classpath, jvmdoc_targets, create_jvmdoc_command)
          else:
            self._generate_individual(classpath, jvmdoc_targets, create_jvmdoc_command)

      if catalog:
        for target in targets:
          gendir = self._gendir(target)
          jvmdocs = []
          for root, dirs, files in os.walk(gendir):
            jvmdocs.extend(os.path.relpath(os.path.join(root, f), gendir) for f in files)
          self.context.products.get(self._jvmdoc.product_type).add(target, gendir, jvmdocs)

  def _generate_combined(self, classpath, targets, create_jvmdoc_command):
    gendir = os.path.join(self._output_dir, 'combined')
    if targets:
      safe_mkdir(gendir, clean=True)
      command = create_jvmdoc_command(classpath, gendir, *targets)
      if command:
        create_jvmdoc(command, gendir)
    if self.open:
      binary_util.ui_open(os.path.join(gendir, 'index.html'))

  def _generate_individual(self, classpath, targets, create_jvmdoc_command):
    jobs = {}
    for target in targets:
      gendir = self._gendir(target)
      command = create_jvmdoc_command(classpath, gendir, target)
      if command:
        jobs[gendir] = (target, command)

    if jobs:
      with contextlib.closing(
            multiprocessing.Pool(processes=min(len(jobs), multiprocessing.cpu_count()))) as pool:
        # map would be a preferable api here but fails after the 1st batch with an internal:
        # ...
        #  File "...src/python/twitter/pants/tasks/jar_create.py", line 170, in javadocjar
        #      pool.map(createjar, jobs)
        #    File "...lib/python2.6/multiprocessing/pool.py", line 148, in map
        #      return self.map_async(func, iterable, chunksize).get()
        #    File "...lib/python2.6/multiprocessing/pool.py", line 422, in get
        #      raise self._value
        #  NameError: global name 'self' is not defined
        futures = []
        for gendir, (target, command) in jobs.items():
          futures.append(pool.apply_async(create_jvmdoc, args=(command, gendir)))

        for future in futures:
          result, gendir = future.get()
          target, command = jobs[gendir]
          if result != 0:
            message = 'Failed to process %s for %s [%d]: %s' % (
                      self._jvmdoc.tool_name, target, result, command)
            if self.ignore_failure:
              self.context.log.warn(message)
            else:
              raise TaskError(message)

  def _gendir(self, target):
    return os.path.join(self._output_dir, target.id)


def create_jvmdoc(command, gendir):
  safe_mkdir(gendir, clean=True)
  process = subprocess.Popen(command)
  result = process.wait()
  return result, gendir



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from collections import defaultdict

import twitter.pants

from twitter.common.collections import OrderedSet
from twitter.pants import get_buildroot
from twitter.pants.base.target import Target
from twitter.pants.base.build_file import BuildFile
from twitter.pants.targets.sources import SourceRoot
from twitter.pants.tasks import TaskError
from twitter.pants.tasks.console_task import ConsoleTask


class ReverseDepmap(ConsoleTask):
  """Outputs all targets whose dependencies include at least one of the input targets."""

  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    super(ReverseDepmap, cls).setup_parser(option_group, args, mkflag)

    option_group.add_option(mkflag("transitive"), mkflag("transitive", negate=True),
                            dest="reverse_depmap_transitive", default=False,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] List transitive dependees.")

    option_group.add_option(mkflag("closed"), mkflag("closed", negate=True),
                            dest="reverse_depmap_closed", default=False,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] Include the input targets in the output along with "
                                 "the dependees.")

    option_group.add_option(mkflag('type'), dest='dependees_type', action='append', default=[],
                            help="Identifies target types to include. Multiple type inclusions "
                                 "can be specified at once in a comma separated list or else by "
                                 "using multiple instances of this flag.")

  def __init__(self, context):
    ConsoleTask.__init__(self, context)

    self._transitive = context.options.reverse_depmap_transitive
    self._closed = context.options.reverse_depmap_closed
    self._dependees_type = context.options.dependees_type

  def console_output(self, _):
    buildfiles = OrderedSet()
    if self._dependees_type:
      base_paths = OrderedSet()
      for dependees_type in self._dependees_type:
        try:
          # Try to do a fully qualified import 1st for filtering on custom types.
          from_list, module, type_name = dependees_type.rsplit('.', 2)
          __import__('%s.%s' % (from_list, module), fromlist=[from_list])
        except (ImportError, ValueError):
          # Fall back on pants provided target types.
          if hasattr(twitter.pants, dependees_type):
            type_name = getattr(twitter.pants, dependees_type)
          else:
            raise TaskError('Invalid type name: %s' % dependees_type)
        # Find the SourceRoot for the given input type
        base_paths.update(SourceRoot.roots(type_name))
      if not base_paths:
        raise TaskError('No SourceRoot set for any target type in %s.' % self._dependees_type +
                        '\nPlease define a source root in BUILD file as:' +
                        '\n\tsource_root(\'<src-folder>\', %s)' % ', '.join(self._dependees_type))
      for base_path in base_paths:
        buildfiles.update(BuildFile.scan_buildfiles(get_buildroot(), base_path))
    else:
      buildfiles = BuildFile.scan_buildfiles(get_buildroot())

    dependees_by_target = defaultdict(set)
    for buildfile in buildfiles:
      for address in Target.get_all_addresses(buildfile):
        for target in Target.get(address).resolve():
          # TODO(John Sirois): tighten up the notion of targets written down in a BUILD by a
          # user vs. targets created by pants at runtime.
          target = self.get_concrete_target(target)
          if hasattr(target, 'dependencies'):
            for dependencies in target.dependencies:
              for dependency in dependencies.resolve():
                dependency = self.get_concrete_target(dependency)
                dependees_by_target[dependency].add(target)

    roots = set(self.context.target_roots)
    if self._closed:
      for root in roots:
        yield str(root.address)

    for dependant in self.get_dependants(dependees_by_target, roots):
      yield str(dependant.address)

  def get_dependants(self, dependees_by_target, roots):
    check = set(roots)
    known_dependants = set()
    while True:
      dependants = set(known_dependants)
      for target in check:
        dependants.update(dependees_by_target[target])
      check = dependants - known_dependants
      if not check or not self._transitive:
        return dependants - set(roots)
      known_dependants = dependants

  def get_concrete_target(self, target):
    return target.derived_from if isinstance(target, Target) else target



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

import sys

class PantsBootstrapError(ImportError):
  pass

try:
  from mako.template import Template
except ImportError:
  raise PantsBootstrapError("Could not properly bootstrap Pants because Mako is missing!")

import os
import pprint

class TemplateData(dict):
  """Encapsulates data for a mako template as a property-addressable read-only map-like struct."""

  def __init__(self, **kwargs):
    dict.__init__(self, kwargs)

  def extend(self, **kwargs):
    """Returns a new TemplateData with this template's data overlayed by the key value pairs
    specified as keyword arguments."""

    props = self.copy()
    props.update(kwargs)
    return TemplateData(**props)

  def __setattr__(self, key, value):
    raise AttributeError("Mutation not allowed - use %s.extend(%s = %s)" % (self, key, value))

  def __getattr__(self, key):
    if key in self:
      return self[key]
    else:
      return object.__getattribute__(self, key)

  def __str__(self):
    return 'TemplateData(%s)' % pprint.pformat(self)

class Generator(object):
  """Generates pants intermediary output files using a configured mako template."""

  _module_directory = '/tmp/pants-%s' % os.environ['USER']

  def __init__(self, template_text, **template_data):
    self._template = Template(text = template_text,
                              module_directory = Generator._module_directory)
    self.template_data = template_data

  def write(self, stream):
    """Applies the template to the template data and writes the result to the given file-like
    stream."""

    stream.write(self._template.render(**self.template_data))



from __future__ import division

from collections import defaultdict

from twitter.pants import is_concrete
from twitter.pants.tasks.console_task import ConsoleTask


class PageRank(ConsoleTask):
  """Measure how "depended-upon" the targets are."""

  def console_output(self, targets):
    dependencies_by_target = defaultdict(set)
    dependees_by_target = defaultdict(set)
    pagerank_by_target = defaultdict(lambda: 1.0)

    self._calc_deps(targets, dependencies_by_target, dependees_by_target)
    self._pagerank(dependees_by_target, dependencies_by_target, pagerank_by_target)
    return self._report(pagerank_by_target)

  def _calc_deps(self, targets, dependencies_by_target, dependees_by_target):
    for target in filter(lambda x: hasattr(x, "dependencies"), targets):
      if not dependencies_by_target.has_key(target):
        for dependency in target.dependencies:
          for resolved in dependency.resolve():
            if is_concrete(resolved):
              dependencies_by_target[target].add(resolved)

      for dependency in target.dependencies:
        for resolved in dependency.resolve():
          if is_concrete(resolved):
            dependees_by_target[resolved].add(target)

  def _pagerank(self, dependees_by_target, dependencies_by_target, pagerank_by_target):
    """Calculate PageRank."""
    d = 0.85
    for x in range(0, 100):
      for target, dependees in dependees_by_target.iteritems():
        contributions = map(
          lambda t: pagerank_by_target[t] / len(dependencies_by_target[t]), dependees)
        pagerank_by_target[target] = (1-d) + d * sum(contributions)

  def _report(self, pagerank_by_target):
    """Yield the report lines."""
    for target in sorted(pagerank_by_target, key=pagerank_by_target.get, reverse=True):
      yield '%f - %s' % (pagerank_by_target[target], target)



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

"""Support for wholesale archive creation and extraction in a uniform API across archive types."""

import os

from abc import abstractmethod
from zipfile import ZIP_DEFLATED

from twitter.common.collections.ordereddict import OrderedDict
from twitter.common.contextutil import open_tar, open_zip
from twitter.common.lang import AbstractClass


class Archiver(AbstractClass):
  @classmethod
  def extract(cls, path, outdir):
    """Extracts an archive's contents to the specified outdir."""
    raise NotImplementedError()

  @abstractmethod
  def create(self, basedir, outdir, name, prefix=None):
    """Creates an archive of all files found under basedir to a file at outdir of the given name.

    If prefix is specified, it should be prepended to all archive paths.
    """


class TarArchiver(Archiver):
  """An archiver that stores files in a tar file with optional compression."""

  @classmethod
  def extract(cls, path, outdir):
    with open_tar(path) as tar:
      tar.extractall(outdir)

  def __init__(self, mode, extension):
    Archiver.__init__(self)
    self.mode = mode
    self.extension = extension

  def create(self, basedir, outdir, name, prefix=None):
    tarpath = os.path.join(outdir, '%s.%s' % (name, self.extension))
    with open_tar(tarpath, self.mode, dereference=True) as tar:
      tar.add(basedir, arcname=prefix or '')
    return tarpath


class ZipArchiver(Archiver):
  """An archiver that stores files in a zip file with optional compression."""

  @classmethod
  def extract(cls, path, outdir):
    """OS X's python 2.6.1 has a bug in zipfile that makes it unzip directories as regular files.

    This method should work on for python 2.6-3.x.
    """
    with open_zip(path) as zip:
      for path in zip.namelist():
        # While we're at it, we also perform this safety test.
        if path.startswith('/') or path.startswith('..'):
          raise ValueError('Zip file contains unsafe path: %s' % path)
        # Ignore directories. extract() will create parent dirs as needed.
        if not path.endswith('/'):
          zip.extract(path, outdir)

  def __init__(self, compression):
    Archiver.__init__(self)
    self.compression = compression

  def create(self, basedir, outdir, name, prefix=None):
    zippath = os.path.join(outdir, '%s.zip' % name)
    with open_zip(zippath, 'w', compression=ZIP_DEFLATED) as zip:
      for root, _, files in os.walk(basedir):
        for file in files:
          full_path = os.path.join(root, file)
          relpath = os.path.relpath(full_path, basedir)
          if prefix:
            relpath = os.path.join(prefix, relpath)
          zip.write(full_path, relpath)
    return zippath


TAR = TarArchiver('w:', 'tar')
TGZ = TarArchiver('w:gz', 'tar.gz')
TBZ2 = TarArchiver('w:bz2', 'tar.bz2')
ZIP = ZipArchiver(ZIP_DEFLATED)

_ARCHIVER_BY_TYPE = OrderedDict(tar=TAR, tgz=TGZ, tbz2=TBZ2, zip=ZIP)

TYPE_NAMES = frozenset(_ARCHIVER_BY_TYPE.keys())

def archiver(typename):
  """Returns Archivers in common configurations.

  The typename must correspond to one of the following:
  'tar'   Returns a tar archiver that applies no compression and emits .tar files.
  'tgz'   Returns a tar archiver that applies gzip compression and emits .tar.gz files.
  'tbz2'  Returns a tar archiver that applies bzip2 compression and emits .tar.bz2 files.
  'zip'   Returns a zip archiver that applies standard compression and emits .zip files.
  """
  archiver = _ARCHIVER_BY_TYPE.get(typename)
  if not archiver:
    raise ValueError('No archiver registered for %r' % typename)
  return archiver



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import shlex
import subprocess

from twitter.pants.ivy import profile_classpath
from twitter.pants.java import runjava
from twitter.pants.targets import JavaLibrary, JavaTests, ScalaLibrary, ScalaTests

from .jvm_task import JvmTask

from . import Task


def is_jvm(target):
  return isinstance(target, (JavaLibrary, JavaTests, ScalaLibrary, ScalaTests))


class ScalaRepl(JvmTask):
  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    option_group.add_option(mkflag("jvmargs"), dest = "run_jvmargs", action="append",
      help = "Run the repl in a jvm with these extra jvm args.")

    option_group.add_option(mkflag("args"), dest = "run_args", action="append",
      help = "Run the repl with these main() args.")

  def __init__(self, context):
    Task.__init__(self, context)
    self.jvm_args = context.config.getlist('scala-repl', 'jvm_args', default=[])
    if context.options.run_jvmargs:
      for arg in context.options.run_jvmargs:
        self.jvm_args.extend(shlex.split(arg))
    self.args = []
    if context.options.run_args:
      for arg in context.options.run_args:
        self.args.extend(shlex.split(arg))
    self.confs = context.config.getlist('scala-repl', 'confs')
    self.profile = context.config.get('scala-repl', 'profile')
    self.main = context.config.get('scala-repl', 'main')

  def execute(self, targets):
    # The repl session may last a while, allow concurrent pants activity during this pants idle
    # period.
    self.context.lock.release()

    self.save_stty_options()
    try:
      _, classpath = profile_classpath(self.profile)
      runjava(
        classpath=self.classpath(classpath, confs=self.confs),
        main=self.main,
        args=self.args,
        jvm_args=self.jvm_args
      )
    except KeyboardInterrupt:
      # TODO(John Sirois): Confirm with Steve Gury that finally does not work on mac and an
      # explicit catch of KeyboardInterrupt is required.
      pass
    self.restore_ssty_options()

  def save_stty_options(self):
    """
    The scala REPL changes some stty parameters and doesn't save/restore them after
    execution, so if you have a terminal with non-default stty options, you end
    up to a broken terminal (need to do a 'reset').
    """
    self.stty_options = self.run_cmd("stty -g 2>/dev/null")

  def restore_ssty_options(self):
    self.run_cmd("stty " + self.stty_options)

  def run_cmd(self, cmd):
    po = subprocess.Popen(cmd, shell=True, stdout=subprocess.PIPE)
    stdout, _ = po.communicate()
    return stdout



# ==================================================================================================
# Copyright 2012 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os

from zipfile import  ZIP_STORED, ZIP_DEFLATED

from twitter.common.contextutil import temporary_dir
from twitter.common.dirutil import safe_mkdir

from twitter.pants import get_buildroot, get_version, is_internal
from twitter.pants.fs.archive import ZIP
from twitter.pants.java.jar import open_jar, Manifest
from twitter.pants.tasks.jvm_binary_task import JvmBinaryTask


class BinaryCreate(JvmBinaryTask):
  """Creates a runnable monolithic binary deploy jar."""

  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    JvmBinaryTask.setup_parser(option_group, args, mkflag)
    option_group.add_option(mkflag("compressed"), mkflag("compressed", negate=True),
                            dest="binary_create_compressed", default=True,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] Create a compressed binary jar.")

    option_group.add_option(mkflag("zip64"), mkflag("zip64", negate=True),
                            dest="binary_create_zip64", default=False,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] Create the binary jar with zip64 extensions.")

  def __init__(self, context):
    JvmBinaryTask.__init__(self, context)

    self.outdir = os.path.abspath(
      context.options.jvm_binary_create_outdir or
      context.config.get('binary-create', 'outdir',
                         default=context.config.getdefault('pants_distdir'))
    )
    self.compression = ZIP_DEFLATED if context.options.binary_create_compressed else ZIP_STORED
    self.zip64 = (
      context.options.binary_create_zip64
      or context.config.getbool('binary-create', 'zip64', default=False)
    )
    self.deployjar = context.options.jvm_binary_create_deployjar

    context.products.require('jars', predicate=self.is_binary)
    context.products.require('classes', predicate=self.is_binary)
    if self.deployjar:
      self.require_jar_dependencies()

  def execute(self, targets):
    for binary in filter(self.is_binary, targets):
      self.create_binary(binary)

  def create_binary(self, binary):
    import platform
    safe_mkdir(self.outdir)

    jarmap = self.context.products.get('jars')

    binary_jarname = '%s.jar' % binary.basename
    binaryjarpath = os.path.join(self.outdir, binary_jarname)
    self.context.log.info('creating %s' % os.path.relpath(binaryjarpath, get_buildroot()))

    with open_jar(binaryjarpath, 'w', compression=self.compression, allowZip64=self.zip64) as jar:
      if self.deployjar:
        def add_jars(target):
          generated = jarmap.get(target)
          if generated:
            for basedir, jars in generated.items():
              for internaljar in jars:
                self.dump(os.path.join(basedir, internaljar), jar)
        binary.walk(add_jars, is_internal)

        for basedir, externaljar in self.list_jar_dependencies(binary):
          self.dump(os.path.join(basedir, externaljar), jar)

      binary_classes = self.context.products.get('classes').get(binary)
      if binary_classes:
        for prefix, filenames in binary_classes.items():
          for filename in filenames:
            jar.write(os.path.join(prefix, filename), arcname=filename)

      manifest = Manifest()
      manifest.addentry(Manifest.MANIFEST_VERSION, '1.0')
      manifest.addentry(
        Manifest.CREATED_BY,
        'python %s pants %s (Twitter, Inc.)' % (platform.python_version(), get_version())
      )
      main = binary.main or '*** java -jar not supported, please use -cp and pick a main ***'
      manifest.addentry(Manifest.MAIN_CLASS,  main)
      jar.writestr(Manifest.PATH, manifest.contents())

      jarmap.add(binary, self.outdir, [binary_jarname])

  def dump(self, jarpath, jarfile):
    self.context.log.debug('  dumping %s' % jarpath)

    with temporary_dir() as tmpdir:
      ZIP.extract(jarpath, tmpdir)
      for root, dirs, files in os.walk(tmpdir):
        for file in files:
          path = os.path.join(root, file)
          relpath = os.path.relpath(path, tmpdir)
          if Manifest.PATH != relpath:
            jarfile.write(path, relpath)




# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os

from twitter.common.contextutil import open_zip
from twitter.common.dirutil import safe_open
from twitter.common.dirutil import safe_rmtree

from twitter.pants import get_buildroot
from twitter.pants.targets import IdlJvmThriftLibrary
from twitter.pants.tasks import Task, TaskError


class IdlExtract(Task):
  """Extracts idl thrift jars and points the sources to the extracted paths."""

  def _is_idl_jar_thrift_library(self, target):
    return isinstance(target, IdlJvmThriftLibrary)

  def __init__(self, context):
    Task.__init__(self, context)
    context.products.require('idl_dependencies', predicate=self._is_idl_jar_thrift_library)
    self._workdir = self.context.config.get('idl-extract', 'workdir')
    if self._workdir is None:
      self._workdir = os.path.join(self.get_workdir(), 'idl')

  def execute(self, targets):
    depmap = self.context.products.get('idl_dependencies')

    def process(target):
      sources = set()
      # TODO (tina): This should guard against items that haven't changed to prevent re-extraction
      extract_base = os.path.join(self._workdir, target.id)
      safe_rmtree(extract_base)
      target.target_base = os.path.relpath(extract_base, get_buildroot())
      for basedir, jars in depmap.get(target).items():
        if len(jars) != 1:
          raise TaskError(
            "%s must have exactly one jar dependency but found %s" % (target.name, jars))
        sources.update(self._extract(extract_base, os.path.join(basedir, jars[0])))
      target.sources = sources

    for target in targets:
      target.walk(process, predicate=lambda target: isinstance(target, IdlJvmThriftLibrary))

  def _extract(self, extract_base, jar_path):
    self.context.log.debug('Extracting idl jar %s to: %s' % (jar_path, extract_base))

    with open_zip(jar_path) as jar:
      sources = set()
      for path in jar.namelist():
        if path.endswith('.thrift'):
          contents = jar.read(path)
          with safe_open(os.path.join(extract_base, path), 'w') as out:
            out.write(contents)
          sources.add(path)

      self.context.log.debug('Found thrift IDL sources: %s' % sources)
      return sources



# ==================================================================================================
# Copyright 2014 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function


import atexit
import os
import shutil

from twitter.common.dirutil import safe_mkdir, safe_rmtree
from twitter.pants.targets import SourceRoot


class FileCopier(object):
  """
  Copies the source file from registered SourceRoots for input target types to
  the destination root location at relative position.
  """

  class FileNotFoundInSourceRoot(Exception):
    def __init__(self, file, roots):
      Exception.__init__(self,
                         "Source %s not found in root(s): %s"
                         % (file, roots))

  class FileCopierError(Exception):
    def __init__(self, file, roots, error):
      Exception.__init__(self,
                         "Could not copy the source %s in root(s): %s due to %s"
                         % (file, roots, error))


  def __init__(self, copy_root):
    """
    :param copy_root: The destination at which you want to copy the files.
    """
    self._copy_root = copy_root
    self._sources = []
    atexit.register(self.cleanup, os.getpid())

  def find_and_copy_relative_file(self, source, target_types):
    """
    Finds the source file at relative position in the source root registered for
    the input target_types and copies it into the copy root at the same relative postion.
    :param source: Source file to copy
    :param target_types: The source root to look for source file listed in the order
      of preference.
    """
    safe_mkdir(self._copy_root)
    source_roots = []
    for target in target_types:
      source_roots.extend(list(SourceRoot.roots(target)))
    source_roots_owning = self._find_file(source, source_roots)
    if not source_roots_owning:
      raise self.FileNotFoundInSourceRoot(source, source_roots)
    return self.copy_relative_file(source, source_roots_owning)

  def copy_relative_file(self, source, source_root):
    """
    Copies the source file at relative position from given source_root
    to the copy_root at the relative position
    :param source:  Source file at relative position to source_root
    :param source_root: source root   in which the source file exists
    """
    return self._copyfile(source, source_root)

  def _find_file(self, source, bases):
    for base in bases:
      if os.path.commonprefix([os.path.abspath(base),
                               os.path.abspath(source)]) == os.path.abspath(base):
        return base
    return None

  def _copyfile(self, source, source_root):
    target_source = os.path.relpath(source, source_root)
    dest_file = os.path.join(self._copy_root, target_source)
    safe_mkdir(os.path.dirname(dest_file))
    try:
      shutil.copyfile(source, dest_file)
      self._sources.append(dest_file)
      return dest_file
    except IOError as e:
      raise(self.FileCopierError(source, source_root, e.strerror))
    except OSError as e:
      raise(self.FileCopierError(source, source_root, e.strerror))

  def cleanup(self, pid):
    if os.getpid() == pid:
    # Do cleanup - we're in the original parent process that parsed the BUILD files
    # And not in a forked NailgunTask executed much later.
      for source in self._sources:
        target = os.path.join(self._copy_root, source)
        if os.path.exists(target):
          os.unlink(target)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import os
import re
import sys

from twitter.common.dirutil import safe_mkdir, safe_open

from twitter.pants import binary_util, is_java, is_scala, is_test, junit_tests
from twitter.pants.ivy import profile_classpath
from twitter.pants.java import runjava

from .jvm_task import JvmTask

from . import Task, TaskError


_RUNNER_MAIN = 'com.twitter.common.junit.runner.ConsoleRunner'


class JUnitRun(JvmTask):
  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    option_group.add_option(mkflag("skip"), mkflag("skip", negate=True), dest = "junit_run_skip",
                            action="callback", callback=mkflag.set_bool, default=False,
                            help = "[%default] Skip running tests")

    option_group.add_option(mkflag("debug"), mkflag("debug", negate=True), dest = "junit_run_debug",
                            action="callback", callback=mkflag.set_bool, default=False,
                            help = "[%default] Run junit tests with a debugger")

    option_group.add_option(mkflag("fail-fast"), mkflag("fail-fast", negate=True),
                            dest = "junit_run_fail_fast",
                            action="callback", callback=mkflag.set_bool, default=False,
                            help = "[%default] Fail fast on the first test failure in a suite")

    option_group.add_option(mkflag("batch-size"), type = "int", default=sys.maxint,
                            dest = "junit_run_batch_size",
                            help = "[ALL] Runs at most this many tests in a single test process.")

    option_group.add_option(mkflag("jvmargs"), dest = "junit_run_jvmargs", action="append",
                            help = "Runs junit tests in a jvm with these extra jvm args.")

    option_group.add_option(mkflag("test"), dest = "junit_run_tests", action="append",
                            help = "[%default] Force running of just these tests.  Tests can be "
                                   "specified using any of: [classname], [classname]#[methodname], "
                                   "[filename] or [filename]#[methodname]")

    outdir = mkflag("outdir")
    option_group.add_option(outdir, dest="junit_run_outdir",
                            help="Emit output in to this directory.")

    xmlreport = mkflag("xmlreport")
    option_group.add_option(xmlreport, mkflag("xmlreport", negate=True),
                            dest = "junit_run_xmlreport",
                            action="callback", callback=mkflag.set_bool, default=False,
                            help = "[%default] Causes an xml report to be output for each test "
                                   "class that is run.")

    option_group.add_option(mkflag("per-test-timer"), mkflag("per-test-timer", negate=True),
                            dest = "junit_run_per_test_timer",
                            action="callback", callback=mkflag.set_bool, default=False,
                            help = "[%default] Shows progress and timer for each test "
                                   "class that is run.")

    option_group.add_option(mkflag("default-parallel"), mkflag("default-parallel", negate=True),
                            dest = "junit_run_default_parallel",
                            action="callback", callback=mkflag.set_bool, default=False,
                            help = "[%default] Whether to run classes without @TestParallel or "
                                   "@TestSerial annotations in parallel.")

    option_group.add_option(mkflag("parallel-threads"), type = "int", default=0,
                            dest = "junit_run_parallel_threads",
                            help = "Number of threads to run tests in parallel. 0 for autoset.")

    option_group.add_option(mkflag("test-shard"), dest = "junit_run_test_shard",
                            help = "Subset of tests to run, in the form M/N, 0 <= M < N."
                                   "For example, 1/3 means run tests number 2, 5, 8, 11, ...")

    option_group.add_option(mkflag("coverage"), mkflag("coverage", negate=True),
                            dest = "junit_run_coverage",
                            action="callback", callback=mkflag.set_bool, default=False,
                            help = "[%default] Collects code coverage data")

    coverage_patterns = mkflag("coverage-patterns")
    option_group.add_option(coverage_patterns, dest="junit_run_coverage_patterns",
                            action="append",
                            help="By default all non-test code depended on by the selected tests "
                                 "is measured for coverage during the test run.  By specifying "
                                 "coverage patterns you can select which classes and packages "
                                 "should be counted.  Values should be class name prefixes in "
                                 "dotted form with ? and * wildcard support. If preceded with a - "
                                 "the pattern is excluded. "
                                 "For example, to include all code in com.twitter.raven except "
                                 "claws and the eye you would use: "
                                 "%(flag)s=com.twitter.raven.* "
                                 "%(flag)s=-com.twitter.raven.claw "
                                 "%(flag)s=-com.twitter.raven.Eye"
                                 "This option can be specified multiple times. " % dict(
                                    flag=coverage_patterns
                                 ))

    option_group.add_option(mkflag("coverage-console"), mkflag("coverage-console", negate=True),
                            dest = "junit_run_coverage_console",
                            action="callback", callback=mkflag.set_bool, default=True,
                            help = "[%default] Outputs a simple coverage report to the console.")

    option_group.add_option(mkflag("coverage-xml"), mkflag("coverage-xml", negate=True),
                            dest = "junit_run_coverage_xml",
                            action="callback", callback=mkflag.set_bool, default=False,
                            help = "[%%default] Produces an xml coverage report in %s." % outdir)

    coverage_html_flag = mkflag("coverage-html")
    option_group.add_option(coverage_html_flag, mkflag("coverage-html", negate=True),
                            dest = "junit_run_coverage_html",
                            action="callback", callback=mkflag.set_bool, default=False,
                            help = "[%%default] Produces an html coverage report in %s." % outdir)

    option_group.add_option(mkflag("coverage-html-open"), mkflag("coverage-html-open", negate=True),
                            dest = "junit_run_coverage_html_open",
                            action="callback", callback=mkflag.set_bool, default=False,
                            help = "[%%default] Tries to open the generated html coverage report, "
                                   "implies %s." % coverage_html_flag)

    option_group.add_option(mkflag("suppress-output"), mkflag("suppress-output", negate=True),
                            dest = "junit_run_suppress_output",
                            action="callback", callback=mkflag.set_bool, default=True,
                            help = "[%%default] Redirects test output to files in %s.  "
                                   "Implied by %s" % (outdir, xmlreport))

    option_group.add_option(mkflag("arg"), dest="junit_run_arg",
                            action="append",
                            help = "An arbitrary argument to pass directly to the test runner. "
                                   "This option can be specified multiple times.")

  def __init__(self, context):
    Task.__init__(self, context)

    self.confs = context.config.getlist('junit-run', 'confs')
    self.junit_profile = context.config.get('junit-run', 'junit_profile')
    self.emma_profile = context.config.get('junit-run', 'emma_profile')

    self.jvm_args = context.config.getlist('junit-run', 'jvm_args', default=[])
    if context.options.junit_run_jvmargs:
      self.jvm_args.extend(context.options.junit_run_jvmargs)
    if context.options.junit_run_debug:
      self.jvm_args.extend(context.config.getlist('jvm', 'debug_args'))

    self.test_classes = context.options.junit_run_tests
    self.context.products.require('classes')

    self.outdir = (
      context.options.junit_run_outdir
      or context.config.get('junit-run', 'workdir')
    )

    self.batch_size = context.options.junit_run_batch_size
    self.fail_fast = context.options.junit_run_fail_fast

    self.coverage = context.options.junit_run_coverage
    self.coverage_filters = context.options.junit_run_coverage_patterns or []
    self.coverage_dir = os.path.join(self.outdir, 'coverage')
    self.coverage_instrument_dir = os.path.join(self.coverage_dir, 'classes')
    self.coverage_metadata_file = os.path.join(self.coverage_dir, 'coverage.em')
    self.coverage_file = os.path.join(self.coverage_dir, 'coverage.ec')

    self.coverage_report_console = context.options.junit_run_coverage_console
    self.coverage_console_file = os.path.join(self.coverage_dir, 'coverage.txt')

    self.coverage_report_xml = context.options.junit_run_coverage_xml
    self.coverage_xml_file = os.path.join(self.coverage_dir, 'coverage.xml')

    self.coverage_report_html_open = context.options.junit_run_coverage_html_open
    self.coverage_report_html = (
      self.coverage_report_html_open
      or context.options.junit_run_coverage_html
    )
    self.coverage = self.coverage or self.coverage_report_html_open
    self.coverage_html_file = os.path.join(self.coverage_dir, 'html', 'index.html')

    self.opts = []
    if context.options.junit_run_xmlreport or context.options.junit_run_suppress_output:
      if self.fail_fast:
        self.opts.append('-fail-fast')
      if context.options.junit_run_xmlreport:
        self.opts.append('-xmlreport')
      self.opts.append('-suppress-output')
      self.opts.append('-outdir')
      self.opts.append(self.outdir)

    if context.options.junit_run_per_test_timer:
      self.opts.append('-per-test-timer')
    if context.options.junit_run_default_parallel:
      self.opts.append('-default-parallel')
    self.opts.append('-parallel-threads')
    self.opts.append(str(context.options.junit_run_parallel_threads))

    if context.options.junit_run_test_shard:
      self.opts.append('-test-shard')
      self.opts.append(context.options.junit_run_test_shard)

    if context.options.junit_run_arg:
      self.opts.extend(context.options.junit_run_arg)

  def _partition(self, tests):
    stride = min(self.batch_size, len(tests))
    for i in xrange(0, len(tests), stride):
      yield tests[i:i+stride]

  def execute(self, targets):
    if not self.context.options.junit_run_skip:
      tests = list(self.normalize_test_classes() if self.test_classes
                                                 else self.calculate_tests(targets))
      if tests:
        _, classpath = profile_classpath(self.junit_profile)
        junit_classpath = self.classpath(classpath, confs=self.confs)

        def run_tests(classpath, main, jvm_args=None):
          # TODO(John Sirois): Integrated batching with the test runner.  As things stand we get
          # results summaries for example for each batch but no overall summary.
          # http://jira.local.twitter.com/browse/AWESOME-1114
          result = 0
          for batch in self._partition(tests):
            with binary_util.safe_args(batch) as batch_tests:
              result += abs(runjava(
                classpath=classpath,
                main=main,
                args=self.opts + batch_tests,
                jvm_args=(jvm_args or []) + self.jvm_args,
              ))
              if result != 0 and self.fail_fast:
                break
          if result != 0:
            raise TaskError('java %s ... exited non-zero (%i)' % (main, result))

        if self.coverage:
          _, emma_classpath = profile_classpath(self.emma_profile)

          def instrument_code():
            safe_mkdir(self.coverage_instrument_dir, clean=True)
            with binary_util.safe_args(self.get_coverage_patterns(targets)) as patterns:
              args = [
                'instr',
                '-out', self.coverage_metadata_file,
                '-d', self.coverage_instrument_dir,
                '-cp', os.pathsep.join(junit_classpath),
                '-exit'
              ]
              for pattern in patterns:
                args.extend(['-filter', pattern])
              main = 'emma'
              result = runjava(emma_classpath, main, args=args)
              if result != 0:
                raise TaskError("java %s ... exited non-zero (%i)"
                                " 'failed to instrument'" % (main, result))

          def generate_reports():
            args = [
              'report',
              '-in', self.coverage_metadata_file,
              '-in', self.coverage_file,
              '-exit'
            ]
            source_bases = set()
            def collect_source_base(target):
              if self.is_coverage_target(target):
                source_bases.add(target.target_base)
            for target in self.test_target_candidates(targets):
              target.walk(collect_source_base)
            for source_base in source_bases:
              args.extend(['-sp', source_base])

            sorting = ['-Dreport.sort', '+name,+class,+method,+block']
            if self.coverage_report_console:
              args.extend(['-r', 'txt',
                           '-Dreport.txt.out.file=%s' % self.coverage_console_file] + sorting)
            if self.coverage_report_xml:
              args.extend(['-r', 'xml','-Dreport.xml.out.file=%s' % self.coverage_xml_file])
            if self.coverage_report_html:
              args.extend(['-r', 'html',
                           '-Dreport.html.out.file=%s' % self.coverage_html_file,
                           '-Dreport.out.encoding=UTF-8'] + sorting)

            main = 'emma'
            result = runjava(emma_classpath, main, args=args)
            if result != 0:
              raise TaskError("java %s ... exited non-zero (%i)"
                              " 'failed to generate code coverage reports'" % (main, result))

            if self.coverage_report_console:
              with safe_open(self.coverage_console_file) as console_report:
                sys.stdout.write(console_report.read())
            if self.coverage_report_html_open:
              binary_util.ui_open(self.coverage_html_file)

          instrument_code()
          try:
            # Coverage runs over instrumented classes require the instrumented classes come 1st in
            # the classpath followed by the normal classpath.  The instrumentation also adds a
            # dependency on emma libs that must be satisfied on the classpath.
            run_tests([self.coverage_instrument_dir] + junit_classpath + emma_classpath,
                      _RUNNER_MAIN,
                      jvm_args=['-Demma.coverage.out.file=%s' % self.coverage_file])
          finally:
            generate_reports()
        else:
          run_tests(junit_classpath, _RUNNER_MAIN)

  def is_coverage_target(self, tgt):
    return (is_java(tgt) or is_scala(tgt)) and not is_test(tgt) and not tgt.is_codegen

  def get_coverage_patterns(self, targets):
    if self.coverage_filters:
      return self.coverage_filters
    else:
      classes_under_test = set()
      classes_by_source = self.context.products.get('classes')
      def add_sources_under_test(tgt):
        if self.is_coverage_target(tgt):
          for source in tgt.sources:
            classes = classes_by_source.get(source)
            if classes:
              for base, classes in classes.items():
                classes_under_test.update(
                  JUnitRun.classfile_to_classname(cls) for cls in classes
                )

      for target in targets:
        target.walk(add_sources_under_test)
      return classes_under_test

  def normalize_test_classes(self):
    for cls in self.test_classes:
      for c in self.normalize(cls):
        yield c

  def test_target_candidates(self, targets):
    for target in targets:
      if isinstance(target, junit_tests):
        yield target

  def calculate_tests(self, targets):
    for target in self.test_target_candidates(targets):
      for test in target.sources:
        for cls in self.normalize(test, target.target_base):
          yield cls

  @staticmethod
  def classfile_to_classname(cls):
    clsname, _ = os.path.splitext(cls.replace('/', '.'))
    return clsname

  def normalize(self, classname_or_file, basedir=None):
    components = classname_or_file.split('#', 2)
    classname = components[0]
    methodname = '#' + components[1] if len(components) == 2 else ''

    classes_by_source = self.context.products.get('classes')
    def relpath_toclassname(path):
      classes = classes_by_source.get(path)
      if not classes:
        # Its perfectly valid - if questionable - to have a source file with no classes when, for
        # example, the source file has all its code commented out.
        self.context.log.warn('File %s contains no classes' % os.path.join(basedir, path))
      else:
        for base, classes in classes.items():
          for cls in classes:
            yield JUnitRun.classfile_to_classname(cls)

    if basedir:
      for classname in relpath_toclassname(classname):
        yield classname + methodname
    elif os.path.exists(classname):
      basedir = calculate_basedir(classname)
      for classname in relpath_toclassname(os.path.relpath(classname, basedir)):
        yield classname + methodname
    else:
      yield classname + methodname


PACKAGE_PARSER = re.compile(r'^\s*package\s+([\w.]+)\s*;?\s*')


def calculate_basedir(file):
  with open(file, 'r') as source:
    for line in source:
      match = PACKAGE_PARSER.match(line)
      if match:
        package = match.group(1)
        packagedir = package.replace('.', '/')
        dir = os.path.dirname(file)
        if not dir.endswith(packagedir):
          raise TaskError('File %s declares a mismatching package %s' % (file, package))
        return dir[:-len(packagedir)]

  raise TaskError('Could not calculate a base dir for: %s' % file)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import errno
import hashlib
import os
import shutil

from abc import abstractmethod
from collections import namedtuple
from functools import partial

from twitter.common.lang import AbstractClass, Compatibility
from twitter.common.dirutil import safe_mkdir, safe_open, safe_rmtree

from twitter.pants.fs import safe_filename


CacheKey = namedtuple('CacheKey', ['sources', 'hash', 'filename'])


class SourceScope(AbstractClass):
  """Selects sources of a given scope from targets."""

  @staticmethod
  def for_selector(selector):
    class Scope(SourceScope):
      def select(self, target):
        return selector(target)
    return Scope()

  @abstractmethod
  def select(self, target):
    """Selects source files from the given target and returns them as absolute paths."""

  def valid(self, target):
    """Returns True if the given target can be used with this SourceScope."""
    return hasattr(target, 'expand_files')


NO_SOURCES = SourceScope.for_selector(lambda t: ())


def get_target_sources(recursive, include_buildfile, t):
  return t.expand_files(recursive, include_buildfile) if hasattr(t, "expand_files") else []

TARGET_SOURCES = SourceScope.for_selector(partial(get_target_sources, False, False))
TRANSITIVE_SOURCES = SourceScope.for_selector(partial(get_target_sources, True, False))


class BuildCache(object):
  """Caches build artifacts and invalidates based on the SHA1 hash of source files."""

  VERSION = 0

  def __init__(self, root):
    self._root = os.path.join(root, str(BuildCache.VERSION))
    safe_mkdir(self._root)

  def key_for(self, id, sources):
    """Get a cache key representing the given id'd entity and its associated source files."""
    return self._key_for(id, self._sources_hash(sources), sources)

  def key_for_target(self, target, sources=TARGET_SOURCES, fingerprint_extra=None):
    """Get a key representing the given target name and its sources.

    :target: The target to create a CacheKey for.
    :sources: A source scope to select from the target for hashing, defaults to TARGET_SOURCES.
    :fingerprint_extra: A that accepts a sha hash and updates it with extra fingerprint data.
    """
    if (not sources or not sources.valid(target)) and not fingerprint_extra:
      raise ValueError('A target needs to have at least one of sources or a '
                       'fingerprint_extra function to generate a CacheKey.')
    if not sources:
      sources = NO_SOURCES

    srcs = sorted(sources.select(target))
    sha = self._sources_hash(srcs)
    if fingerprint_extra:
      fingerprint_extra(sha)
    return self._key_for(target.id, sha, srcs)

  def _key_for(self, id, sha, sources):
    filename = os.path.join(self._root, id)
    return CacheKey(sources, sha.hexdigest(), filename)

  def invalidate(self, cache_key):
    """Invalidates this cache key and any cached files associated with it.

    :param cache_key: A CacheKey object (as returned by BuildCache.key_for().
    """
    safe_rmtree(cache_key.filename)
    sha_file = self._sha_file(cache_key)
    if os.path.exists(sha_file):
      os.unlink(sha_file)

  def needs_update(self, cache_key):
    """Check if the given target is cached.

    :param cache_key: A CacheKey object (as returned by BuildCache.key_for().
    :returns: True if the cached version of the target is out of date.
    """
    cached_sha = self._read_sha(cache_key)
    return cached_sha != cache_key.hash

  def update(self, cache_key, build_artifacts=None, artifact_root=None):
    """Cache the output of a build.

    If the cache area contains an existing object with the same (path, source_sha) its path will
    be returned. If no such object exists, builder will be called with a path inside the staging
    area and should create a new object.

    :param cache_key: A CacheKey object (typically returned by BuildCache.key_for().
    :param build_artifacts: List of paths to generated artifacts under artifact_root.
    :param artifact_root: Optional root directory under which artifacts are stored.
    """
    safe_rmtree(cache_key.filename)
    for artifact in build_artifacts or ():
      rel_path = os.path.basename(artifact) \
          if artifact_root is None \
          else os.path.relpath(artifact, artifact_root)
      assert not rel_path.startswith('..'), \
        'Weird: artifact=%s, rel_path=%s' % (artifact, rel_path)
      artifact_dest = os.path.join(cache_key.filename, rel_path)
      dir_name = os.path.dirname(artifact_dest)
      if not os.path.exists(dir_name):
        os.makedirs(dir_name)
      if os.path.isdir(artifact):
        shutil.copytree(artifact, artifact_dest)
      else:
        shutil.copy(artifact, artifact_dest)
    self._write_sha(cache_key)

  def use_cached_files(self, cache_key, copy_fn):
    """Use cached files, typically by hard-linking them from the cache into a staging area.

    :param cache_key: A CacheKey object (typically returned by BuildCache.key_for().
    :param copy_fn: A function with the signature copy_fn(absolute_src_path, relative_dst_path) that
        will copy cached files into the desired destination.
    """
    for relative_filename, filename in self._walk_paths([cache_key.filename]):
      copy_fn(filename, relative_filename)

  def _walk_paths(self, paths):
    """Recursively walk the given paths.

    :returns: Iterable of (relative_path, absolute_path).
    """
    assert not isinstance(paths, Compatibility.string)
    for path in sorted(paths):
      if os.path.isdir(path):
        for dir_name, _, filenames in sorted(os.walk(path)):
          for filename in filenames:
            filename = os.path.join(dir_name, filename)
            yield os.path.relpath(filename, path), filename
      else:
        yield os.path.basename(path), path

  def _sources_hash(self, paths):
    """Generate SHA1 digest from the content of all files under the given paths."""
    sha = hashlib.sha1()

    for relative_filename, filename in self._walk_paths(paths):
      with open(filename, "rb") as fd:
        sha.update(Compatibility.to_bytes(relative_filename))
        sha.update(fd.read())

    return sha

  def _key(self, key):
    return key.replace(os.path.sep, '.')

  def _sha_file(self, cache_key):
    return os.path.join(self._root, safe_filename(os.path.basename(cache_key.filename), '.hash'))

  def _write_sha(self, cache_key):
    with safe_open(self._sha_file(cache_key), 'w') as fd:
      fd.write(cache_key.hash)

  def _read_sha(self, cache_key):
    try:
      with open(self._sha_file(cache_key), 'rb') as fd:
        return fd.read().strip()
    except IOError as e:
      if e.errno != errno.ENOENT:
        raise



# ==================================================================================================
# Copyright 2013 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from collections import defaultdict, namedtuple

from twitter.common import log
from twitter.common.collections import OrderedDict, OrderedSet, maybe_list

from twitter.pants import TaskError, is_internal
from twitter.pants.goal import Goal
from twitter.pants.targets.internal import InternalTarget

from .engine import Engine


class GroupMember(namedtuple('GroupMember', ['group', 'name', 'predicate'])):
  """Represents a member of a goal group."""

  @classmethod
  def from_goal(cls, goal):
    """Creates a ``GroupMember`` from goal metadata."""
    if not isinstance(goal, Goal):
      raise ValueError('The given goal must be a ``Goal`` object, given %s' % goal)
    if not goal.group:
      raise ValueError('Can only form a GroupMember from goals with a group defined, goal %s '
                       'has no group' % goal.name)
    return cls(goal.group.name, goal.name, goal.group.predicate)


class GroupIterator(object):
  """Iterates the goals in a group over the chunks they own,"""

  def __init__(self, targets, group_members):
    """Creates an iterator that yields tuples of ``(GroupMember, [chunk Targets])``.

    Chunks will be returned least dependant to most dependant such that a group member processing a
    chunk can be assured that any dependencies of the chunk have been processed already.

    :param list targets: The universe of targets to divide up amongst group members.
    :param list group_members: A list of group members that forms the group to iterate.
    """
    # TODO(John Sirois): These validations should be happening sooner in the goal registration
    # process.
    assert len(map(lambda m: m.group, group_members)) != 1, 'Expected a single group'
    assert len(map(lambda m: m.name, group_members)) == len(group_members), (
      'Expected group members with unique names')

    self._targets = maybe_list(targets, expected_type=InternalTarget, raise_type=ValueError)
    self._group_members = group_members

  def __iter__(self):
    for chunk in self._create_chunks():
      for group_member in self._group_members:
        member_chunk = filter(group_member.predicate, chunk)
        if len(member_chunk) > 0:
          yield group_member, member_chunk

  def _create_chunks(self):
    def discriminator(tgt):
      for group_member in self._group_members:
        if group_member.predicate(tgt):
          return group_member.name
      return None

    # TODO(John Sirois): coalescing should be made available in another spot, InternalTarget is jvm
    # specific, and all we care is that the Targets have dependencies defined
    coalesced = InternalTarget.coalesce_targets(self._targets, discriminator)
    coalesced = list(reversed(coalesced))

    chunks = []
    flavor = None
    chunk_start = 0
    for chunk_num, target in enumerate(coalesced):
      target_flavor = discriminator(target)
      if target_flavor != flavor and chunk_num > chunk_start:
        chunks.append((flavor, OrderedSet(coalesced[chunk_start:chunk_num])))
        chunk_start = chunk_num
      flavor = target_flavor
    if chunk_start < len(coalesced):
      chunks.append((flavor, OrderedSet(coalesced[chunk_start:])))

    log.debug('::: created chunks(%d)' % len(chunks))
    for i, (flavor, chunk) in enumerate(chunks):
      log.debug('  chunk(%d:%s):\n\t%s'
                % (i, flavor, '\n\t'.join(sorted(map(str, chunk)))))

    return map(lambda (flavor, chunk): chunk, chunks)


class GroupEngine(Engine):
  """The classical phase engine that has direct knowledge of groups and the bang algorithm.

  For grouped goals this engine attempts to make as few passes as possible through the target groups
  found.
  """

  class PhaseExecutor(object):
    def __init__(self, context, phase, tasks_by_goal):
      self._context = context
      self._phase = phase
      self._tasks_by_goal = tasks_by_goal

    @property
    def phase(self):
      return self._phase

    def attempt(self, timer, explain):
      """Executes the named phase against the current context tracking goal executions in executed.
      """

      def execute_task(goal, task, targets):
        """Execute and time a single goal that has had all of its dependencies satisfied."""
        with timer.timed(goal):
          # TODO (Senthil Kumaran):
          # Possible refactoring of the Task Execution Logic (AWESOME-1019)
          if explain:
            self._context.log.debug("Skipping execution of %s in explain mode" % goal.name)
          else:
            task.execute(targets)

      goals = self._phase.goals()
      if not goals:
        raise TaskError('No goals installed for phase %s' % self._phase)

      run_queue = []
      goals_by_group = {}
      for goal in goals:
        if goal.group:
          group_name = goal.group.name
          if group_name not in goals_by_group:
            group_goals = [goal]
            run_queue.append((group_name, group_goals))
            goals_by_group[group_name] = group_goals
          else:
            goals_by_group[group_name].append(goal)
        else:
          run_queue.append((None, [goal]))

      # OrderedSet takes care of not repeating chunked task execution mentions
      execution_phases = defaultdict(OrderedSet)

      for group_name, goals in run_queue:
        if not group_name:
          goal = goals[0]
          self._context.log.info('[%s:%s]' % (self._phase, goal.name))
          execution_phases[self._phase].add(goal.name)
          execute_task(goal, self._tasks_by_goal[goal], self._context.targets())
        else:
          goals_by_group_member = OrderedDict((GroupMember.from_goal(g), g) for g in goals)
          # TODO(Travis Crawford): Targets should be filtered by is_concrete rather than
          # is_internal, however, at this time python targets are not internal targets.
          chunks = GroupIterator(self._context.targets(is_internal), goals_by_group_member.keys())
          for group_member, goal_chunk in chunks:
            goal = goals_by_group_member[group_member]
            self._context.log.info('[%s:%s:%s]' % (self._phase, group_name, goal.name))
            execution_phases[self._phase].add((group_name, goal.name))
            execute_task(goal, self._tasks_by_goal[goal], goal_chunk)

      if explain:
        tasks_by_goalname = dict((goal.name, task.__class__.__name__)
                                 for goal, task in self._tasks_by_goal.items())

        def expand_goal(goal):
          if len(goal) == 2:  # goal is (group, goal)
            group_name, goal_name = goal
            task_name = tasks_by_goalname[goal_name]
            return "%s:%s->%s" % (group_name, goal_name, task_name)
          else:
            task_name = tasks_by_goalname[goal]
            return "%s->%s" % (goal, task_name)

        for phase, goals in execution_phases.items():
          goal_to_task = ", ".join(expand_goal(goal) for goal in goals)
          print("%s [%s]" % (phase, goal_to_task))

  @classmethod
  def _prepare(cls, context, phases):
    tasks_by_goal = {}

    # We loop here because a prepared goal may introduce new BUILDs and thereby new Goals/Phases.
    # We need to prepare these in a subsequent loop until the set of phases and goals quiesces.
    prepared_goals = set()
    round_num = 0
    while True:
      phases = list(cls.execution_order(phases))
      if prepared_goals == reduce(lambda goals, p: goals | set(p.goals()), phases, set()):
        break

      round_num += 1
      context.log.debug('Preparing goals in round %d' % round_num)
      # Prepare tasks roots to leaves and allow for downstream tasks requiring products from
      # upstream tasks they depend upon.
      for phase in reversed(phases):
        for goal in reversed(phase.goals()):
          if goal not in prepared_goals:
            context.log.debug('preparing: %s:%s' % (phase.name, goal.name))
            prepared_goals.add(goal)
            task = goal.task_type(context)
            tasks_by_goal[goal] = task

    return map(lambda p: cls.PhaseExecutor(context, p, tasks_by_goal), phases)

  def attempt(self, timer, context, phases):
    phase_executors = self._prepare(context, phases)

    execution_phases = ' -> '.join(map(str, map(lambda e: e.phase.name, phase_executors)))
    context.log.debug('Executing goals in phases %s' % execution_phases)

    explain = getattr(context.options, 'explain', None)
    if explain:
      print("Phase Execution Order:\n\n%s\n" % execution_phases)
      print("Phase [Goal->Task] Order:\n")

    for phase_executor in phase_executors:
      phase_executor.attempt(timer, explain)



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

import collections
import copy
import os
import sys

from contextlib import contextmanager

from twitter.common.lang import Compatibility

from twitter.pants import get_buildroot
from twitter.pants.base import BuildFile, Config


class ParseContext(object):
  """Defines the context of a parseable BUILD file target and provides a mechanism for targets to
  discover their context when invoked via eval.
  """

  class ContextError(Exception):
    """Indicates an action that requires a BUILD file parse context was attempted outside any."""

  _active = collections.deque([])
  _parsed = set()

  @classmethod
  def locate(cls):
    """Attempts to find the current root directory and buildfile.

    If there is an active parse context (see do_in_context), then it is returned.
    """
    if not ParseContext._active:
      raise cls.ContextError('No parse context active.')
    return ParseContext._active[-1]

  @staticmethod
  @contextmanager
  def temp(basedir=None):
    """Activates a temporary parse context in the given basedir relative to the build root or else
    in the build root dir itself if no basedir is specified.
    """
    context = ParseContext(BuildFile(get_buildroot(), basedir or 'BUILD.temp', must_exist=False))
    with ParseContext.activate(context):
      yield

  @classmethod
  @contextmanager
  def activate(cls, ctx):
    """Activates the given ParseContext."""
    if hasattr(ctx, '_on_context_exit'):
      raise cls.ContextError('Context actions registered outside this parse context arg active')

    try:
      ParseContext._active.append(ctx)
      ctx._on_context_exit = []
      yield
    finally:
      for func, args, kwargs in ctx._on_context_exit:
        func(*args, **kwargs)
      del ctx._on_context_exit
      ParseContext._active.pop()

  def __init__(self, buildfile):
    self.buildfile = buildfile
    self._active_buildfile = buildfile
    self._parsed = False

  @staticmethod
  def default_globals(config=None):
    """
    Has twitter.pants.*, but not file-specfic things like __file__
    If you want to add new imports to be available to all BUILD files, add a section to the config similar to:
    [parse]
    headers: ['from test import get_jar',]

    You may also need to add new roots to the sys.path. see _run in pants_exe.py
    """
    pants_context = {}
    ast1 = compile("from twitter.pants import *", "<string>", "exec")
    ast2 = compile("from twitter.common.quantity import Amount, Time", "<string>", "exec")
    Compatibility.exec_function(ast1, pants_context)
    Compatibility.exec_function(ast2, pants_context)

    # TODO: This can be replaced once extensions are enabled with https://github.com/pantsbuild/pants/issues/5
    if config:
      headers = config.getlist('parse', 'headers', default=[])
      for exec_file in headers:
        ast1 = compile(exec_file, "<string>", "exec")
        Compatibility.exec_function(ast1, pants_context)

    return pants_context

  def parse(self, **globalargs):
    """The entry point to parsing of a BUILD file.

    Changes the working directory to the BUILD file directory and then evaluates the BUILD file
    with the ROOT_DIR and __file__ globals set in addition to any globals specified as kwargs.  As
    target methods are parsed they can examine the stack to find these globals and thus locate
    themselves for the purposes of finding files.

    See locate().
    """
    if self.buildfile not in ParseContext._parsed:
      buildfile_family = tuple(self.buildfile.family())

      pants_context = self.default_globals(Config.load())

      with ParseContext.activate(self):
        start = os.path.abspath(os.curdir)
        try:
          os.chdir(self.buildfile.parent_path)
          for buildfile in buildfile_family:
            self._active_buildfile = buildfile
            # We may have traversed a sibling already, guard against re-parsing it.
            if buildfile not in ParseContext._parsed:
              ParseContext._parsed.add(buildfile)

              eval_globals = copy.copy(pants_context)
              eval_globals.update({
                  'ROOT_DIR': buildfile.root_dir,
                  '__file__': buildfile.full_path,
              })
              eval_globals.update(globalargs)
              Compatibility.exec_function(buildfile.code(), eval_globals)
        finally:
          os.chdir(start)

  def on_context_exit(self, func, *args, **kwargs):
    """ Registers a command to invoke just before this parse context is exited.

    It is an error to attempt to register an on_context_exit action outside an active parse
    context.
    """
    if not hasattr(self, '_on_context_exit'):
      raise self.ContextError('Can only register context exit actions when a parse context '
                              'is active')

    if not callable(func):
      raise TypeError('func must be a callable object')

    self._on_context_exit.append((func, args, kwargs))

  def do_in_context(self, work):
    """Executes the callable work in this parse context."""
    if not callable(work):
      raise TypeError('work must be a callable object')

    with ParseContext.activate(self):
      return work()

  def __repr__(self):
    return '%s(%s)' % (type(self).__name__, self.buildfile)

  @property
  def current_buildfile(self):
    """ This property return the current build file being parsed from all BUILD files co-located
    with this BUILD file within the family.
    """
    return self._active_buildfile



# ==================================================================================================
# Copyright 2011 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

from __future__ import print_function

import copy
import functools
import getpass
import hashlib
import logging
import os
import pkgutil
import shutil
import sys

from collections import defaultdict

from twitter.common.collections import OrderedDict, OrderedSet
from twitter.common.config import Properties
from twitter.common.dirutil import safe_open, safe_rmtree
from twitter.common.log.options import LogOptions

from twitter.pants import (
    get_buildroot,
    get_scm,
    is_codegen,
    is_exported as provides,
    is_internal,
    is_java,
    is_scala)
from twitter.pants.base import Address, Target
from twitter.pants.base.generator import Generator, TemplateData
from twitter.pants.ivy import Bootstrapper, Ivy
from twitter.pants.targets import (
    AnnotationProcessor,
    InternalTarget,
    JavaLibrary,
    JavaThriftLibrary,
    Resources,
    ScalaLibrary,
    ThriftJar,
    ThriftLibrary)
from twitter.pants.tasks.scm_publish import ScmPublish, Semver

from . import Task, TaskError


class PushDb(object):
  @staticmethod
  def load(path):
    """Loads a pushdb maintained in a properties file at the given path."""
    with open(path, 'r') as props:
      properties = Properties.load(props)
      return PushDb(properties)

  def __init__(self, props):
    self._props = props

  def as_jar_with_version(self, target):
    """
      Given an internal target, return a JarDependency with the last published revision filled in.
    """
    jar_dep, db_get, _ = self._accessors_for_target(target)

    major = int(db_get('revision.major', '0'))
    minor = int(db_get('revision.minor', '0'))
    patch = int(db_get('revision.patch', '0'))
    snapshot = db_get('revision.snapshot', 'false').lower() == 'true'
    sha = db_get('revision.sha', None)
    fingerprint = db_get('revision.fingerprint', None)
    semver = Semver(major, minor, patch, snapshot=snapshot)
    jar_dep.rev = semver.version()
    return jar_dep, semver, sha, fingerprint

  def set_version(self, target, version, sha, fingerprint):
    version = version if isinstance(version, Semver) else Semver.parse(version)
    _, _, db_set = self._accessors_for_target(target)
    db_set('revision.major', version.major)
    db_set('revision.minor', version.minor)
    db_set('revision.patch', version.patch)
    db_set('revision.snapshot', str(version.snapshot).lower())
    db_set('revision.sha', sha)
    db_set('revision.fingerprint', fingerprint)

  def _accessors_for_target(self, target):
    jar_dep, _, exported = target.get_artifact_info()
    if not exported:
      raise ValueError

    def key(prefix):
      return '%s.%s%%%s' % (prefix, jar_dep.org, jar_dep.name)

    def getter(prefix, default=None):
      return self._props.get(key(prefix), default)

    def setter(prefix, value):
      self._props[key(prefix)] = value

    return jar_dep, getter, setter

  def dump(self, path):
    """Saves the pushdb as a properties file to the given path."""
    with open(path, 'w') as props:
      Properties.dump(self._props, props)


class DependencyWriter(object):
  """
    Builds up a template data representing a target and applies this to a template to produce a
    dependency descriptor.
  """

  def __init__(self, get_db, template_relpath):
    self.get_db = get_db
    self.template_relpath = template_relpath

  def write(self, target, path, confs=None, synth=False):
    def as_jar(internal_target, is_tgt=False):
      jar, _, _, _ = self.get_db(internal_target).as_jar_with_version(internal_target)
      if synth and is_tgt:
        jar.name += '-only'
      return jar

    # TODO(John Sirois): a dict is used here to de-dup codegen targets which have both the original
    # codegen target - say java_thrift_library - and the synthetic generated target (java_library)
    # Consider reworking codegen tasks to add removal of the original codegen targets when rewriting
    # the graph
    dependencies = OrderedDict()
    internal_codegen = {}
    for dep in target_internal_dependencies(target):
      jar = as_jar(dep)
      dependencies[(jar.org, jar.name)] = self.internaldep(jar, dep, synth)
      if is_codegen(dep):
        internal_codegen[jar.name] = jar.name
    for jar in target.jar_dependencies:
      if jar.rev:
        classifier = jar.classifier if isinstance(jar, ThriftJar) else None
        dependencies[(jar.org, jar.name)] = self.jardep(jar, classifier=classifier)
    target_jar = self.internaldep(as_jar(target, is_tgt=True)).extend(
      dependencies=dependencies.values()
    )

    template_kwargs = self.templateargs(target_jar, confs, synth)
    with safe_open(path, 'w') as output:
      template = pkgutil.get_data(__name__, self.template_relpath)
      Generator(template, **template_kwargs).write(output)

  def templateargs(self, target_jar, confs=None, synth=False):
    """
      Subclasses must return a dict for use by their template given the target jar template data
      and optional specific ivy configurations.
    """
    raise NotImplementedError()

  def internaldep(self, jar_dependency, dep=None, synth=False):
    """
      Subclasses must return a template data for the given internal target (provided in jar
      dependency form).
    """
    raise NotImplementedError()

  def jardep(self, jar_dependency, classifier=None):
    """Subclasses must return a template data for the given external jar dependency."""
    raise NotImplementedError()

  def create_exclude(self, exclude):
    return TemplateData(org=exclude.org, name=exclude.name)


class PomWriter(DependencyWriter):
  def __init__(self, get_db):
    super(PomWriter, self).__init__(get_db, os.path.join('templates', 'jar_publish', 'pom.mk'))

  def templateargs(self, target_jar, confs=None, synth=False):
    return dict(artifact=target_jar)

  def jardep(self, jar, classifier=None, synth=False):
    return TemplateData(
      org=jar.org,
      name=jar.name + ('-only' if synth else ''),
      rev=jar.rev,
      scope='compile',
      classifier=(classifier if classifier is not None else jar.classifier),
      excludes=[self.create_exclude(exclude) for exclude in jar.excludes if exclude.name]
    )

  def internaldep(self, jar_dependency, dep=None, synth=False):
    classifier = 'idl' if (is_codegen(dep) and synth) or isinstance(dep, ThriftLibrary) else None
    return self.jardep(jar_dependency, classifier=classifier, synth=synth)


class IvyWriter(DependencyWriter):
  def __init__(self, get_db):
    super(IvyWriter, self).__init__(get_db, os.path.join('templates', 'ivy_resolve', 'ivy.mk'))

  def templateargs(self, target_jar, confs=None, synth=False):
    return dict(lib=target_jar.extend(
      is_idl=synth,
      publications=set(confs) if confs else set(),
      overrides=None
    ))

  def _jardep(self, jar, transitive=True, configurations='default', classifier=None, synth=False):
    return TemplateData(
      org=jar.org,
      module=jar.name + ('-only' if synth else ''),
      version=jar.rev,
      mutable=False,
      force=jar.force,
      excludes=[self.create_exclude(exclude) for exclude in jar.excludes],
      transitive=transitive,
      artifacts=jar.artifacts,
      is_idl=(classifier == 'idl'),
      configurations=configurations,
    )

  def jardep(self, jar, classifier=None):
    return self._jardep(jar,
      transitive=jar.transitive,
      configurations=';'.join(jar._configurations),
      classifier=classifier
    )

  def internaldep(self, jar_dependency, dep=None, synth=False):
    classifier = 'idl' if (is_codegen(dep) and synth) or isinstance(dep, ThriftLibrary) else None
    return self._jardep(jar_dependency, classifier=classifier, synth=synth)


def is_exported(target):
  return provides(target) and isinstance(target, (AnnotationProcessor, JavaLibrary, ScalaLibrary,
                                                  ThriftLibrary))


def coordinate(org, name, rev=None):
  return '%s#%s;%s' % (org, name, rev) if rev else '%s#%s' % (org, name)


def jar_coordinate(jar, rev=None):
  return coordinate(jar.org, jar.name, rev or jar.rev)


def target_internal_dependencies(target):
  return filter(lambda tgt: not isinstance(tgt, Resources), target.internal_dependencies)


class JarPublish(ScmPublish, Task):
  """Publish jars to a maven repository.

  At a high-level, pants uses `Apache Ivy <http://ant.apache.org/ivy/>`_ to
  publish artifacts to Maven-style repositories. Pants performs prerequisite
  tasks like compiling, creating jars, and generating ``pom.xml`` files then
  invokes Ivy to actually publish the artifacts, so publishing is largely
  configured in ``ivysettings.xml``. ``BUILD`` and ``pants.ini`` files
  primarily provide linkage between publishable targets and the
  Ivy ``resolvers`` used to publish them.

  The following target types are publishable: :ref:`bdict_java_library`,
  :ref:`bdict_scala_library`, :ref:`bdict_thrift_library`,
  :ref:`bdict_annotation_processor`.
  Targets to publish and their dependencies must be publishable target
  types and specify the ``provides`` argument. One exception is
  :ref:`bdict_jar`\s - pants will generate a pom file that
  depends on the already-published jar.

  Example usage: ::

     # By default pants will perform a dry-run.
     ./pants goal clean-all publish src/java/com/twitter/mybird

     # Actually publish.
     ./pants goal clean-all publish src/java/com/twitter/mybird --no-publish-dryrun

  Please see ``./pants goal publish -h`` for a detailed description of all
  publishing options.

  Publishing can be configured in ``pants.ini`` as follows.

  ``jar-publish`` section:

  * ``repos`` - Required dictionary of settings for repos that may be pushed to.
  * ``ivy_jvmargs`` - Optional list of JVM command-line args when invoking Ivy.
  * ``restrict_push_branches`` - Optional list of branches to restrict publishing to.

  Example pants.ini jar-publish repos dictionary: ::

     repos = {
       # repository target name is paired with this key
       'myrepo': {
         # ivysettings.xml resolver to use for publishing
         'resolver': 'maven.twttr.com',
         # ivy configurations to publish
         'confs': ['default', 'sources', 'docs'],
         # address of a Credentials target to use when publishing
         'auth': 'address/of/credentials/BUILD:target',
         # help message if unable to initialize the Credentials target.
         'help': 'Please check your credentials and try again.',
       },
     }

  Additionally the ``ivy`` section ``ivy_settings`` property specifies which
  Ivy settings file to use when publishing is required.
  """

  _CONFIG_SECTION = 'jar-publish'

  @classmethod
  def setup_parser(cls, option_group, args, mkflag):
    # TODO(John Sirois): Support a preview mode that outputs a file with entries like:
    # artifact id:
    # revision:
    # publish: (true|false)
    # changelog:
    #
    # Allow re-running this goal with the file as input to support forcing an arbitrary set of
    # revisions and supply of hand endited changelogs.

    option_group.add_option(mkflag("dryrun"), mkflag("dryrun", negate=True),
                            dest="jar_publish_dryrun", default=True,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] Runs through a push without actually pushing "
                                 "artifacts, editing publish dbs or otherwise writing data")

    option_group.add_option(mkflag("commit", negate=True),
                            dest="jar_publish_commit", default=True,
                            action="callback", callback=mkflag.set_bool,
                            help="Turns off commits of the push db for local testing.")

    local_flag = mkflag("local")
    option_group.add_option(local_flag, dest="jar_publish_local",
                            help="Publishes jars to a maven repository on the local filesystem at "
                                 "the specified path.")

    option_group.add_option(mkflag("local-snapshot"), mkflag("local-snapshot", negate=True),
                            dest="jar_publish_local_snapshot", default=True,
                            action="callback", callback=mkflag.set_bool,
                            help="[%%default] If %s is specified, publishes jars with '-SNAPSHOT' "
                                 "revisions." % local_flag)

    option_group.add_option(mkflag("transitive"), mkflag("transitive", negate=True),
                            dest="jar_publish_transitive", default=True,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] Publishes the specified targets and all their "
                                 "internal dependencies transitively.")

    option_group.add_option(mkflag("force"), mkflag("force", negate=True),
                            dest="jar_publish_force", default=False,
                            action="callback", callback=mkflag.set_bool,
                            help="[%default] Forces pushing jars even if there have been no "
                                 "changes since the last push.")

    flag = mkflag('override')
    option_group.add_option(flag, action='append', dest='jar_publish_override',
                            help='''Specifies a published jar revision override in the form:
                            ([org]#[name]|[target spec])=[new revision]

                            For example, to specify 2 overrides:
                            %(flag)s=com.twitter.common#quantity=0.1.2 \\
                            %(flag)s=src/java/com/twitter/common/base=1.0.0 \\
                            ''' % dict(flag=flag))

    flag = mkflag("restart-at")
    option_group.add_option(flag, dest="jar_publish_restart_at",
                            help='''Restart a fail push at the given jar.  Jars can be identified by
                            maven coordinate [org]#[name] or target.

                            For example:
                            %(flag)s=com.twitter.common#quantity

                            Or:
                            %(flag)s=src/java/com/twitter/common/base
                            ''' % dict(flag=flag))

  def __init__(self, context, scm=None):
    Task.__init__(self, context)
    ScmPublish.__init__(self, scm or get_scm(),
                        self.context.config.getlist(
                          JarPublish._CONFIG_SECTION, 'restrict_push_branches'))
    self.outdir = os.path.join(context.config.getdefault('pants_workdir'), 'publish')
    self.cachedir = os.path.join(self.outdir, 'cache')

    self._jvmargs = context.config.getlist(JarPublish._CONFIG_SECTION, 'ivy_jvmargs', default=[])

    if context.options.jar_publish_local:
      local_repo = dict(
        resolver='publish_local',
        path=os.path.abspath(os.path.expanduser(context.options.jar_publish_local)),
        confs=['*'],
        auth=None
      )
      self.repos = defaultdict(lambda: local_repo)
      self.commit = False
      self.snapshot = context.options.jar_publish_local_snapshot
    else:
      self.repos = context.config.getdict(JarPublish._CONFIG_SECTION, 'repos')
      for repo, data in self.repos.items():
        auth = data.get('auth')
        if auth:
          credentials = context.resolve(auth).next()
          user = credentials.username(data['resolver'])
          password = credentials.password(data['resolver'])
          self.context.log.debug('Found auth for repo=%s user=%s' % (repo, user))
          self.repos[repo]['username'] = user
          self.repos[repo]['password'] = password
      self.commit = context.options.jar_publish_commit
      self.snapshot = False

    self.ivycp = context.config.getlist('ivy', 'classpath')
    self.ivysettings = context.config.get('jar-publish', 'ivy_settings')

    self.dryrun = context.options.jar_publish_dryrun
    self.transitive = context.options.jar_publish_transitive
    self.force = context.options.jar_publish_force

    def parse_jarcoordinate(coordinate):
      components = coordinate.split('#', 1)
      if len(components) == 2:
        org, name = components
        return org, name
      else:
        try:
          address = Address.parse(get_buildroot(), coordinate)
          try:
            target = Target.get(address)
            if not target:
              siblings = Target.get_all_addresses(address.buildfile)
              prompt = 'did you mean' if len(siblings) == 1 else 'maybe you meant one of these'
              raise TaskError('%s => %s?:\n    %s' % (address, prompt,
                                                      '\n    '.join(str(a) for a in siblings)))
            if not is_exported(target):
              raise TaskError('%s is not an exported target' % coordinate)
            return target.provides.org, target.provides.name
          except (ImportError, SyntaxError, TypeError):
            raise TaskError('Failed to parse %s' % address.buildfile.relpath)
        except IOError:
          raise TaskError('No BUILD file could be found at %s' % coordinate)

    self.overrides = {}
    if context.options.jar_publish_override:
      def parse_override(override):
        try:
          coordinate, rev = override.split('=', 1)
          try:
            rev = Semver.parse(rev)
          except ValueError as e:
            raise TaskError('Invalid version %s: %s' % (rev, e))
          return parse_jarcoordinate(coordinate), rev
        except ValueError:
          raise TaskError('Invalid override: %s' % override)

      self.overrides.update(parse_override(o) for o in context.options.jar_publish_override)

    self.restart_at = None
    if context.options.jar_publish_restart_at:
      self.restart_at = parse_jarcoordinate(context.options.jar_publish_restart_at)

    context.products.require('jars')
    context.products.require('source_jars')
    context.products.require('idl_jars')

  def execute(self, targets):
    self.check_clean_master(commit=(not self.dryrun and self.commit))

    exported_targets = self.exported_targets()
    self.check_targets(exported_targets)

    pushdbs = {}

    def get_db(tgt):
      # TODO(tdesai) Handle resource type in get_db.
      if tgt.provides is None:
        raise TaskError('trying to publish target %r which does not provide an artifact' % tgt)
      dbfile = tgt.provides.repo.push_db
      result = pushdbs.get(dbfile)
      if not result:
        db = PushDb.load(dbfile)
        repo = self.repos[tgt.provides.repo.name]
        result = (db, dbfile, repo)
        pushdbs[dbfile] = result
      return result

    def get_pushdb(tgt):
      return get_db(tgt)[0]

    def fingerprint_internal(tgt):
      if not is_internal(tgt):
        raise ValueError('Expected an internal target for fingerprinting, got %s' % tgt)
      pushdb, _, _ = get_db(tgt)
      _, _, _, fingerprint = pushdb.as_jar_with_version(tgt)
      return fingerprint or '0.0.0'

    def lookup_synthetic_target(tgt):
      # TODO(phom) this only works for Thrift Library, not Protobuf
      return tgt.derived_from if isinstance(tgt.derived_from, JavaThriftLibrary) else None

    def artifact_path(jar, version, name=None, suffix='', extension='jar', artifact_ext=''):
      return os.path.join(self.outdir, jar.org, jar.name + artifact_ext,
                          '%s%s-%s%s.%s' % ((name or jar.name),
                                            artifact_ext if name != 'ivy' else '',
                                            version,
                                            suffix,
                                            extension))

    def stage_artifact(tgt, jar, version, changelog, confs=None, artifact_ext='', synth=False):
      def path(name=None, suffix='', extension='jar'):
        return artifact_path(jar, version, name=name, suffix=suffix, extension=extension,
                             artifact_ext=artifact_ext)

      with safe_open(path(suffix='-CHANGELOG', extension='txt'), 'w') as changelog_file:
        changelog_file.write(changelog)
      ivyxml = path(name='ivy', extension='xml')

      IvyWriter(get_pushdb).write(tgt, ivyxml, confs=confs, synth=synth)
      PomWriter(get_pushdb).write(tgt, path(extension='pom'), synth=synth)

      return ivyxml

    def copy_artifact(tgt, version, typename, suffix='', artifact_ext=''):
      genmap = self.context.products.get(typename)
      for basedir, jars in genmap.get(tgt).items():
        for artifact in jars:
          path = artifact_path(jar, version, suffix=suffix, artifact_ext=artifact_ext)
          shutil.copy(os.path.join(basedir, artifact), path)

    def stage_artifacts(tgt, jar, version, changelog, confs=None, synth_target=None):
      is_idl = isinstance(tgt, ThriftLibrary)
      class_target = None if is_idl else tgt
      idl_target = synth_target or (tgt if is_idl else None)

      class_ivyxml_path = idl_ivyxml_path = None
      if class_target:
        class_ivyxml_path = stage_artifact(tgt, jar, version, changelog, confs)
        copy_artifact(tgt, version, typename='jars')
        copy_artifact(tgt, version, typename='source_jars', suffix='-sources')

        jarmap = self.context.products.get('javadoc_jars')
        if not jarmap.empty() and (is_java(tgt) or is_scala(tgt)):
          copy_artifact(tgt, version, typename='javadoc_jars', suffix='-javadoc')

      if idl_target:
        synth = bool(synth_target)
        artifact_ext = '-only' if synth else ''
        idl_ivyxml_path = stage_artifact(idl_target, jar, version, changelog, confs=['idl'],
                                         artifact_ext=artifact_ext, synth=synth)
        copy_artifact(idl_target, version, typename='idl_jars', suffix='-idl',
                      artifact_ext=artifact_ext)

      return class_ivyxml_path, idl_ivyxml_path

    if self.overrides:
      print('Publishing with revision overrides:\n  %s' % '\n  '.join(
        '%s=%s' % (coordinate(org, name), rev) for (org, name), rev in self.overrides.items()
      ))

    head_sha = self.scm.commit_id

    safe_rmtree(self.outdir)
    published = []
    skip = (self.restart_at is not None)
    for target in exported_targets:
      synth_target = lookup_synthetic_target(target)
      pushdb, dbfile, repo = get_db(target)
      jar, semver, sha, fingerprint = pushdb.as_jar_with_version(target)

      if synth_target:
        # add idl artifact to the published cache
        tmp_jar = copy.copy(jar)
        tmp_jar.name += '-only'
        published.append(tmp_jar)
      published.append(jar)

      if skip and (jar.org, jar.name) == self.restart_at:
        skip = False

      newver = self.overrides.get((jar.org, jar.name)) or semver.bump()
      if self.snapshot:
        newver = newver.make_snapshot()

      if newver <= semver:
        raise TaskError('Requested version %s must be greater than the current version %s' % (
            newver.version(), semver.version()
            ))

      newfingerprint = self.fingerprint(target, fingerprint_internal)
      no_changes = newfingerprint == fingerprint

      if no_changes:
        changelog = 'No changes for %s - forced push.\n' % jar_coordinate(jar, semver.version())
      else:
        changelog = self.changelog(target, sha) or 'Direct dependencies changed.\n'

      if no_changes and not self.force:
        print('No changes for %s' % jar_coordinate(jar, semver.version()))
        stage_artifacts(target, jar, (newver if self.force else semver).version(), changelog,
                        synth_target=synth_target)
      elif skip:
        print('Skipping %s to resume at %s' % (
          jar_coordinate(jar, (newver if self.force else semver).version()),
          coordinate(self.restart_at[0], self.restart_at[1])
        ))
        stage_artifacts(target, jar, semver.version(), changelog, synth_target=synth_target)
      else:
        if not self.dryrun:
          # Confirm push looks good
          if no_changes:
            print(changelog)
          else:
            print('\nChanges for %s since %s @ %s:\n\n%s' % (
              coordinate(jar.org, jar.name), semver.version(), sha, changelog
            ))
          if os.isatty(sys.stdin.fileno()):
            push = raw_input('Publish %s with revision %s ? [y|N] ' % (
              coordinate(jar.org, jar.name), newver.version()
            ))
            print('\n')
            if push.strip().lower() != 'y':
              raise TaskError('User aborted push')

        pushdb.set_version(target, newver, head_sha, newfingerprint)

        ivyxml, idl_ivyxml = stage_artifacts(target, jar, newver.version(), changelog,
                                             confs=repo['confs'], synth_target=synth_target)

        if self.dryrun:
          print('Skipping publish of %s in test mode.' % jar_coordinate(jar, newver.version()))
        else:
          resolver = repo['resolver']
          path = repo.get('path')

          # Get authentication for the publish repo if needed
          jvm_args = self._jvmargs
          if repo.get('auth'):
            user = repo.get('username')
            password = repo.get('password')
            if user and password:
              jvm_args.append('-Dlogin=%s' % user)
              jvm_args.append('-Dpassword=%s' % password)
            else:
              raise TaskError('Unable to publish to %s. %s' %
                              (repo['resolver'], repo.get('help', '')))

          # Do the publish
          def publish(ivyxml_path):
            ivysettings = self.generate_ivysettings(published, publish_local=path)
            args = [
              '-settings', ivysettings,
              '-ivy', ivyxml_path,
              '-deliverto', '%s/[organisation]/[module]/ivy-[revision].xml' % self.outdir,
              '-publish', resolver,
              '-publishpattern', '%s/[organisation]/[module]/'
                                 '[artifact]-[revision](-[classifier]).[ext]' % self.outdir,
              '-revision', newver.version(),
              '-m2compatible',
            ]

            if LogOptions.stderr_log_level() == logging.DEBUG:
              args.append('-verbose')

            if self.snapshot:
              args.append('-overwrite')

            try:
              ivy = Bootstrapper.default_ivy()
              ivy.execute(args, jvm_args=jvm_args)
            except (Bootstrapper.Error, Ivy.Error) as e:
              raise TaskError('Failed to push %s! %s' % (jar_coordinate(jar, newver.version()), e))

          if ivyxml:
            publish(ivyxml)
          if idl_ivyxml:
            publish(idl_ivyxml)

          if self.commit:
            org = jar.org
            name = jar.name
            rev = newver.version()
            args = dict(
              org=org,
              name=name,
              rev=rev,
              coordinate=coordinate(org, name, rev),
              user=getpass.getuser(),
              cause='with forced revision' if (org, name) in self.overrides else '(autoinc)'
            )

            pushdb.dump(dbfile)
            self.commit_push(coordinate(org, name, rev))
            self.scm.refresh()
            self.scm.tag('%(org)s-%(name)s-%(rev)s' % args,
                         message='Publish of %(coordinate)s initiated by %(user)s %(cause)s' % args)

  def check_targets(self, targets):
    invalid = defaultdict(set)

    def collect(publish_target, walked_target):
      if hasattr(walked_target, "sources") and not walked_target.sources:
        invalid[publish_target].add((walked_target, 'No sources.'))
      if hasattr(walked_target, "provides") and not walked_target.provides:
        invalid[publish_target].add((walked_target, 'Does not provide an artifact.'))

    for target in targets:
      target.walk(functools.partial(collect, target))

    if invalid:
      msg = list()
      for target, reasons in sorted(invalid.items(), reverse=True):
        msg.append('\n  Cannot publish %s due to:' % target.address)
        for invalid_target, reason in sorted(reasons, reverse=True):
          msg.append('\n    %s - %s' % (invalid_target.address, reason))

      raise TaskError('The following errors must be resolved to publish.%s' % ''.join(msg))

  def exported_targets(self):
    candidates = set()
    if self.transitive:
      candidates.update(self.context.targets())
    else:
      candidates.update(self.context.target_roots)

      def get_synthetic(lang, target):
        mappings = self.context.products.get(lang).get(target)
        if mappings:
          for generated in mappings.itervalues():
            for synthetic in generated:
              yield synthetic

      # Handle the case where a code gen target is in the listed roots and the thus the publishable
      # target is a synthetic twin generated by a code gen task upstream.
      for candidate in self.context.target_roots:
        candidates.update(get_synthetic('java', candidate))
        candidates.update(get_synthetic('scala', candidate))

    def exportable(tgt):
      return tgt in candidates and is_exported(tgt)

    return OrderedSet(filter(exportable,
                             reversed(InternalTarget.sort_targets(filter(exportable, candidates)))))

  def fingerprint(self, target, fingerprint_internal):
    sha = hashlib.sha1()

    for source in sorted(target.sources):
      path = os.path.join(target.target_base, source)
      with open(path) as fd:
        sha.update(source)
        sha.update(fd.read())

    # TODO(John Sirois): handle resources and circular dep scala_library java_sources

    for jarsig in sorted([jar_coordinate(j) for j in target.jar_dependencies if j.rev]):
      sha.update(jarsig)

    # TODO(tdesai) Handle resource type in get_db.
    internal_dependencies = sorted(target_internal_dependencies(target), key=lambda t: t.id)
    for internal_target in internal_dependencies:
      fingerprint = fingerprint_internal(internal_target)
      sha.update(fingerprint)

    return sha.hexdigest()

  def changelog(self, target, sha):
    return self.scm.changelog(from_commit=sha,
                              files=[os.path.join(target.target_base, source)
                                     for source in target.sources])

  def generate_ivysettings(self, publishedjars, publish_local=None):
    template_relpath = os.path.join('templates', 'jar_publish', 'ivysettings.mk')
    template = pkgutil.get_data(__name__, template_relpath)
    with safe_open(os.path.join(self.outdir, 'ivysettings.xml'), 'w') as wrapper:
      generator = Generator(template,
                            ivysettings=self.ivysettings,
                            dir=self.outdir,
                            cachedir=self.cachedir,
                            published=[TemplateData(org=jar.org, name=jar.name)
                                       for jar in publishedjars],
                            publish_local=publish_local)
      generator.write(wrapper)
      return wrapper.name
